import os
import numpy as np
import numpy.testing as npt

from cesium import data_management
from cesium.features import graphs
from cesium.features.tests.util import generate_features


# Fixed set of features w/ known values
SCIENCE_FEATS = graphs.GENERAL_FEATS + graphs.LOMB_SCARGLE_FEATS


def test_feature_generation():
    """Compare generated features to reference values."""
    this_dir = os.path.join(os.path.dirname(__file__))
    test_files = [
            os.path.join(this_dir, 'data/257141.dat'),
            os.path.join(this_dir, 'data/245486.dat'),
            os.path.join(this_dir, 'data/247327.dat'),
            ]
    features_extracted = None
    values_computed = None
    for i, ts_data_file_path in enumerate(test_files):
        t, m, e = data_management.parse_ts_data(ts_data_file_path)
        features = generate_features(t, m, e, SCIENCE_FEATS)
        sorted_features = sorted(features.items())
        if features_extracted is None:
            features_extracted = [f[0] for f in sorted_features]
            values_computed = np.zeros((len(test_files),
                len(features_extracted)))
        values_computed[i,:] = [f[1] for f in sorted_features]

    def features_from_csv(filename):
        with open(filename) as f:
            feature_names = f.readline().strip().split(",")
            feature_values = np.loadtxt(f, delimiter=',')

        return feature_names, feature_values

    this_dir = os.path.join(os.path.dirname(__file__))
    features_expected, values_expected = features_from_csv(
        os.path.join(this_dir, "data/expected_features.csv"))

    npt.assert_equal(features_extracted, features_expected)
    npt.assert_array_almost_equal(values_computed, values_expected)



import hashlib
import os
import tarfile

import pandas as pd

from .. import util

try:
    import urllib.request as request
except:
    import urllib2 as request


DATA_PATH = os.path.expanduser('~/.local/')


def _md5sum_file(path):
    """Calculate the MD5 sum of a file."""
    with open(path, 'rb') as f:
        m = hashlib.md5()
        while True:
            data = f.read(8192)
            if not data:
                break
            m.update(data)
    return m.hexdigest()


def download_file(data_dir, base_url, filename):
    """Download a single file into the given directory.

    Parameters
    ----------
    data_dir : str
        Path to directory in which to save file.
    base_url : str
        URL of file to download, minus the file name.
    filename : str
        Name of file to be downloaded.

    Returns
    -------
    str
        The path to the newly downloaded file.
    """
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    file_path = os.path.join(data_dir, filename)
    opener = request.urlopen(base_url + filename)
    with open(file_path, 'wb') as f:
        f.write(opener.read())
    return file_path


def download_and_extract_archives(data_dir, base_url, filenames, md5sums=None,
                                 remove_archive=True):
    """Download list of data archives, verify md5 checksums (if applicable),
    and extract into the given directory.

    Parameters
    ----------
    data_dir : str
        Path to directory in which to download and extract archives.
    base_url : str
        URL of files to download, minus the file names.
    filenames : list or tuple of str
        Name of file to be downloaded.
    md5sums : dict, optional
        Dictionary whose keys are file names and values are
        corresponding hexadecimal md5 checksums to be checked against.
    remove_archive : bool, optional
        Boolean indicating whether to delete the archive(s) from disk
        after the contents have been extracted. Defaults to True.

    Returns
    -------
    list of str
        The paths to the newly downloaded and unzipped files.
    """
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)

    all_file_paths = []
    for fname in filenames:
        archive_path = os.path.join(data_dir, fname)
        opener = request.urlopen(base_url + fname)
        with open(archive_path, 'wb') as f:
            f.write(opener.read())
        if md5sums:
            if _md5sum_file(archive_path) != md5sums[fname]:
                raise ValueError("File {} checksum verification has failed."
                                 " Dataset fetching aborted.".format(fname))
        with util.extract_time_series(archive_path,
                                      cleanup_archive=remove_archive,
                                      extract_dir=data_dir) as file_paths:
            all_file_paths.extend(file_paths)
    return all_file_paths


def build_time_series_archive(archive_path, ts_paths):
    """Write a .tar.gz archive containing the given time series files, as
    required for data uploaded via the front end.

    Parameters
    ----------
    archive_path : str
        Path at which to create the tarfile.
    ts_paths : list of str
        Paths to time-series file to be included in tarfile.
    """
    with tarfile.TarFile(archive_path, 'w') as t:
        for fname in ts_paths:
            t.add(fname, arcname=os.path.basename(fname))


def write_header(header_path, filenames, classes, metadata={}):
    """Write a header file for the given time series files, as required for
    data uploaded via the front end.

    Parameters
    ----------
    header_path : str
        Path at which header file will be created.
    filenames : list of str
        List of time-series file names associated with header file.
    classes : list of str
        List of class names associated with each time-series file.
    metadata : dict, optional
        Dictionary describing meta features associated with each time-series.
        Keys are time-series file names.
    """
    data_dict = {'filename': [util.shorten_fname(f) for f in filenames],
                 'class': classes}
    data_dict.update(metadata)
    df = pd.DataFrame(data_dict)
    df.to_csv(header_path, index=False)



import numpy as np


def stetson_mean(x, weight=100., alpha=2., beta=2., tol=1.e-6, nmax=20):
    """An iteratively weighted mean used in the Stetson variability index"""
    mu = np.median(x)
    for i in range(nmax):
        resid = x - mu
        resid_err = np.abs(resid) * np.sqrt(weight)
        weight1 = weight / (1. + (resid_err / alpha)**beta)
        weight1 /= weight1.mean()
        diff = np.mean(x * weight1) - mu
        mu += diff
        if (np.abs(diff) < tol*np.abs(mu) or np.abs(diff) < tol):
            break

    return mu


def stetson_j(x, y=[], dx=0.1, dy=0.1):
    """
    Robust covariance statistic between pairs of observations x,y
    whose uncertainties are dx,dy. If y is not given, calculates a robust
    variance for x.
    """
    n = len(x)
    x0 = stetson_mean(x, 1./dx**2)
    delta_x = np.sqrt(n / (n - 1.)) * (x - x0) / dx

    if (len(y) > 0):
        y0 = stetson_mean(y, 1./dy**2)
        delta_y = np.sqrt(n / (n - 1.)) * (y - y0) / dy
        p_k = delta_x * delta_y
    else:
        p_k = delta_x**2 - 1.

    return np.mean(np.sign(p_k) * np.sqrt(np.abs(p_k)))


def stetson_k(x, dx=0.1):
    """A robust kurtosis statistic."""
    n = len(x)
    x0 = stetson_mean(x, 1./dx**2)
    delta_x = np.sqrt(n / (n - 1.)) * (x - x0) / dx
    return 1. / 0.798 * np.mean(np.abs(delta_x)) / np.sqrt(np.mean(delta_x**2))



import logging
import os

import numpy as np
from sklearn.externals import joblib

from .. import util
from . import util as dsutil


BASE_URL = "https://github.com/cesium-ml/cesium-data/raw/master/andrzejak/"
ZIP_FILES = ["Z.zip", "O.zip", "N.zip", "F.zip", "S.zip"]
MD5SUMS = {
    "Z.zip": "ca5c761d62704c4d2465822e2131f868",
    "O.zip": "666ade7e9d519935103404d4a8d81d7d",
    "N.zip": "0bb8e39ae7530ba17f55b5b4f14e6a02",
    "F.zip": "10f78c004122c609e8eef74de8790af3",
    "S.zip": "1d560ac1e03a5c19bb7f336e270ff286"
}
CACHE_NAME = "andrzejak.pkl"
ARCHIVE_NAME = "andrzejak.tar.gz"
T_MAX = 23.6

logger = logging.getLogger(__name__)


def download_andrzejak(data_dir):
    """Download sample EEG data.

    Three files are created within `data_dir`:
        - andrzejak.tar.gz (containing .dat files for each time series)
        - andrzejak.csv (header file containing class labels)
        - andrzejak.pkl (cached data for faster loading from disk)

    Parameters
    ----------
    data_dir : str
        Path where downloaded data should be stored.

    Returns
    -------
    dict
        Dictionary with attributes:
            - times: list of (4096,) arrays of time values
            - measurements: list of (4096,) arrays of measurement values
            - classes: array of class labels for each time series
            - archive: path to data archive
            - header: path to header file
    """
    logger.warning("Downloading data from {}".format(BASE_URL))

    ts_paths = dsutil.download_and_extract_archives(data_dir, BASE_URL,
                                                    ZIP_FILES, MD5SUMS)

    # Reformat time series files and add to `andrzejak.tar.gz` archive
    times = []
    measurements = []
    new_ts_paths = []
    classes = []
    for fname in ts_paths:
        m = np.loadtxt(fname)
        t = np.linspace(start=0, stop=T_MAX, num=len(m))
        new_fname = fname[:-4] + '.dat'
        np.savetxt(new_fname, np.vstack((t, m)).T, delimiter=',')
        measurements.append(m)
        times.append(t)
        new_ts_paths.append(new_fname)
        classes.append(os.path.basename(new_fname)[0])  # 'Z001.txt' -> 'Z'
    archive_path = os.path.join(data_dir, ARCHIVE_NAME)
    dsutil.build_time_series_archive(archive_path, new_ts_paths)

    header_path = os.path.join(data_dir, "andrzejak.csv")
    dsutil.write_header(header_path, new_ts_paths, classes)

    extract_dir = os.path.join(data_dir, os.path.basename(ARCHIVE_NAME))
    util.remove_files(ts_paths + new_ts_paths)

    cache_path = os.path.join(data_dir, CACHE_NAME)
    data = dict(times=times, measurements=measurements,
                classes=np.array(classes), archive=archive_path,
                header=header_path)
    joblib.dump(data, cache_path, compress=3)
    return data


def fetch_andrzejak(data_dir=None):
    """Download (if not already downloaded) and load an example EEG dataset.

    Parameters
    ----------
    data_dir : str, optional
        Path where downloaded data should be stored. Defaults to a subdirectory
        `datasets/andrzejak` within `dsutil.DATA_PATH`.

    Returns
    -------
    dict
        Dictionary with attributes:
            - times: list of (4096,) arrays of time values
            - measurements: list of (4096,) arrays of measurement values
            - classes: array of class labels for each time series
            - archive: path to data archive
            - header: path to header file

    References
    ----------
    Andrzejak, Ralph G., et al. "Indications of nonlinear deterministic and
    finite-dimensional structures in time series of brain electrical activity:
    Dependence on recording region and brain state." *Phys. Rev. E* 64.6
    (2001): 061907.
    """

    if data_dir is None:
        data_dir = os.path.join(dsutil.DATA_PATH, "datasets/andrzejak")
    cache_path = os.path.join(data_dir, CACHE_NAME)

    try:
        data = joblib.load(cache_path)
        logger.warning("Loaded data from cached archive.")
    except (ValueError, IOError): #  missing or incompatible cache
        data = download_andrzejak(data_dir)
    return data



import copy
from collections import Iterable
import numpy as np
import pandas as pd
import dask
import dask.threaded
from dask import delayed
from dask.local import reraise
from dask.threaded import pack_exception
from dask.optimization import cull
from sklearn.impute import SimpleImputer as Imputer

from . import time_series
from .time_series import TimeSeries
from .features import generate_dask_graph

__all__ = ['featurize_time_series', 'featurize_single_ts',
           'featurize_ts_files', 'assemble_featureset']


def featurize_single_ts(ts, features_to_use, custom_script_path=None,
                        custom_functions=None, raise_exceptions=True):
    """Compute feature values for a given single time-series. Data is
    returned as dictionaries/lists of lists.

    Parameters
    ----------
    ts : TimeSeries object
        Single time series to be featurized.
    features_to_use : list of str
        List of feature names to be generated.
    custom_functions : dict, optional
        Dictionary of custom feature functions to be evaluated for the given
        time series, or a dictionary representing a dask graph of function
        evaluations. Dictionaries of functions should have keys `feature_name`
        and values functions that take arguments (t, m, e); in the case of a
        dask graph, these arrays should be referenced as 't', 'm', 'e',
        respectively, and any values with keys present in `features_to_use`
        will be computed.
    raise_exceptions : bool, optional
        If True, exceptions during feature computation are raised immediately;
        if False, exceptions are supressed and `np.nan` is returned for the
        given feature and any dependent features. Defaults to True.

    Returns
    -------
    dict
        Dictionary with feature names as keys, lists of feature values (one per
        channel) as values.
    """
    # Initialize empty feature array for all channels
    feature_values = np.empty((len(features_to_use), ts.n_channels))
    for (t_i, m_i, e_i), i in zip(ts.channels(), range(ts.n_channels)):
        feature_graph = generate_dask_graph(t_i, m_i, e_i)
        feature_graph.update(ts.meta_features)

        if custom_functions:
            # If values in custom_functions are functions, add calls to graph
            if all(hasattr(v, '__call__') for v in custom_functions.values()):
                feature_graph.update({feat: f(t_i, m_i, e_i)
                                      for feat, f in custom_functions.items()})
            # Otherwise, custom_functions is another dask graph
            else:
                feature_graph.update(custom_functions)

        # Do not execute in parallel; parallelization has already taken place
        # at the level of time series, so we compute features for a single time
        # series in serial.
        if raise_exceptions:
            raise_callback = reraise
        else:
            raise_callback = lambda e, tb: None
        culled_feature_graph, _ = cull(feature_graph, features_to_use)
        dask_values = dask.get(culled_feature_graph, features_to_use,
                               raise_exception=raise_callback,
                               pack_exception=pack_exception)
        feature_values[:, i] = [x if not isinstance(x, Exception) else np.nan
                                for x in dask_values]
    index = pd.MultiIndex.from_product((features_to_use, range(ts.n_channels)),
                                       names=('feature', 'channel'))
    return pd.Series(feature_values.ravel(), index=index)


def assemble_featureset(features_list, time_series=None,
                        meta_features_list=None, names=None):
    """Transforms raw feature data (as returned by `featurize_single_ts`) into
    a pd.DataFrame.

    Parameters
    ----------
    features_list : list of pd.Series
        List of series (one per time series file) with (feature name, channel)
        multiindex.
    time_series : list of TimeSeries
        If provided, the name and metafeatures from the time series objects
        will be used, overriding the `meta_features_list` and `names` values.
    meta_features_list : list of dict
        If provided, the columns of `metadata` will be added to the featureset.
    names : list of str
        If provided, the (row) index of the featureset will be set accordingly.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns containing feature values, indexed by name.
    """
    if time_series is not None:
        meta_features_list, names = zip(*[(ts.meta_features, ts.name)
                                          for ts in time_series])
    if len(features_list) > 0:
        feat_df = pd.concat(features_list, axis=1, ignore_index=True).T
        feat_df.index = names
    else:
        feat_df = pd.DataFrame(index=names)

    if meta_features_list and any(meta_features_list):  # not all empty dicts
        meta_df = pd.DataFrame(list(meta_features_list), index=names)
        meta_df.columns = pd.MultiIndex.from_tuples([(c, '') for c in meta_df],
                                                    names=['feature', 'channel'])
        feat_df = pd.concat((feat_df, meta_df), axis=1)

    return feat_df


# TODO should this be changed to use TimeSeries objects? or maybe an optional
# argument for TimeSeries? some redundancy here...
def featurize_time_series(times, values, errors=None, features_to_use=[],
                          meta_features={}, names=None,
                          custom_script_path=None, custom_functions=None,
                          scheduler=dask.threaded.get, raise_exceptions=True):
    """Versatile feature generation function for one or more time series.

    For a single time series, inputs may have the form:

    - `times`:  (n,) array or (p, n) array (for p channels of measurement)
    - `values`: (n,) array or (p, n) array (for p channels of measurement)
    - `errors`: (n,) array or (p, n) array (for p channels of measurement)

    For multiple time series, inputs may have the form:

    - `times`: list of (n,) arrays, list of (p, n) arrays (for p channels of
      measurement), or list of lists of (n,) arrays (for
      multichannel data with different time values per channel)
    - `values`: list of (n,) arrays, list of (p, n) arrays (for p channels of
      measurement), or list of lists of (n,) arrays (for
      multichannel data with different time values per channel)
    - `errors`: list of (n,) arrays, list of (p, n) arrays (for p channels of
      measurement), or list of lists of (n,) arrays (for
      multichannel data with different time values per channel)

    In the case of multichannel measurements, each channel will be
    featurized separately, and the index of the output featureset will contain
    a `channel` coordinate.

    Parameters
    ----------
    times : array, list of array, or list of lists of array
        Array containing time values for a single time series, or a list of
        arrays each containing time values for a single time series, or a list
        of lists of arrays for multichannel data with different time values per
        channel
    values : array or list of array
        Array containing measurement values for a single time series, or a list
        of arrays each containing (possibly multivariate) measurement values
        for a single time series, or a list of lists of arrays for multichannel
        data with different time values per channel
    errors : array or list/tuple of array, optional
        Array containing measurement error values for a single time series, or
        a list of arrays each containing (possibly multivariate) measurement
        values for a single time series, or a list of lists of arrays for
        multichannel data with different time values per channel
    features_to_use : list of str, optional
        List of feature names to be generated. Defaults to an empty list, which
        will result in only meta_features features being stored.
    meta_features : dict/Pandas.Series or list of dicts/Pandas.DataFrame
        dict/Series (for a single time series) or DataFrame (for multiple time
        series) of metafeature information; features are added to the output
        featureset, and their values are consumable by custom feature scripts.
    names : str or list of str, optional
        Name or list of names for each time series, if applicable; will be
        stored in the (row) index of the featureset.
    custom_script_path : str, optional
        Path to Python script containing function definitions for the
        generation of any custom features. Defaults to None.
    custom_functions : dict, optional
        Dictionary of custom feature functions to be evaluated for the given
        time series, or a dictionary representing a dask graph of function
        evaluations.  Dictionaries of functions should have keys `feature_name`
        and values functions that take arguments (t, m, e); in the case of a
        dask graph, these arrays should be referenced as 't', 'm', 'e',
        respectively, and any values with keys present in `features_to_use`
        will be computed.
    scheduler : function, optional
        `dask` scheduler function used to perform feature extraction
        computation. Defaults to `dask.threaded.get`.
    raise_exceptions : bool, optional
        If True, exceptions during feature computation are raised immediately;
        if False, exceptions are supressed and `np.nan` is returned for the
        given feature and any dependent features. Defaults to True.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns containing feature values, indexed by name.
    """
    if times is None:
        times = copy.deepcopy(values)
        if isinstance(times, np.ndarray) and (times.ndim == 1
                                              or 1 in times.shape):
            times[:] = np.linspace(0., time_series.DEFAULT_MAX_TIME,
                                   times.size)
        else:
            for t in times:
                if isinstance(t, np.ndarray) and (t.ndim == 1 or 1 in t.shape):
                    t[:] = np.linspace(0., time_series.DEFAULT_MAX_TIME,
                                       t.size)
                else:
                    for t_i in t:
                        t_i[:] = np.linspace(0., time_series.DEFAULT_MAX_TIME,
                                             t_i.size)

    if errors is None:
        errors = copy.deepcopy(values)
        if isinstance(errors, np.ndarray) and (errors.ndim == 1
                                               or 1 in errors.shape):
            errors[:] = time_series.DEFAULT_ERROR_VALUE
        else:
            for e in errors:
                if isinstance(e, np.ndarray) and (e.ndim == 1 or 1 in e.shape):
                    e[:] = time_series.DEFAULT_ERROR_VALUE
                else:
                    for e_i in e:
                        e_i[:] = time_series.DEFAULT_ERROR_VALUE

    # One single-channel time series:
    if not isinstance(values[0], Iterable):
        times, values, errors = [times], [values], [errors]
    # One multi-channel time series:
    elif isinstance(values, np.ndarray) and values.ndim == 2:
        times, values, errors = [times], [values], [errors]

    if names is None:
        names = np.arange(len(times))

    if isinstance(meta_features, pd.Series):
        meta_features = meta_features.to_dict()
    meta_features = pd.DataFrame(meta_features, index=names)

    all_time_series = [delayed(TimeSeries(t, m, e,
                                          meta_features=meta_features.loc[name],
                                          name=name), pure=True)
                       for t, m, e, name in zip(times, values, errors, names)]

    all_features = [delayed(featurize_single_ts, pure=True)(ts, features_to_use,
                                                            custom_script_path,
                                                            custom_functions,
                                                            raise_exceptions)
                    for ts in all_time_series]
    result = delayed(assemble_featureset, pure=True)(all_features, all_time_series)
    return result.compute(scheduler=scheduler)


def featurize_ts_files(ts_paths, features_to_use, custom_script_path=None,
                       custom_functions=None, scheduler=dask.threaded.get,
                       raise_exceptions=True):
    """Feature generation function for on-disk time series (.npz) files.

    By default, computes features concurrently using the
    `dask.threaded.get` scheduler. Other possible options include
    `dask.local.get` for synchronous computation (e.g., when debugging),
    or `dask.distributed.Executor.get` for distributed computation.

    In the case of multichannel measurements, each channel will be
    featurized separately, and the index of the output featureset will contain
    a `channel` coordinate.

    Parameters
    ----------
    ts_paths : list of str
        List of paths to time series data, stored in `numpy` .npz format.
        See `time_series.load` for details.
    features_to_use : list of str, optional
        List of feature names to be generated. Defaults to an empty list, which
        will result in only meta_features features being stored.
    custom_script_path : str, optional
        Path to Python script containing function definitions for the
        generation of any custom features. Defaults to None.
    custom_functions : dict, optional
        Dictionary of custom feature functions to be evaluated for the given
        time series, or a dictionary representing a dask graph of function
        evaluations.  Dictionaries of functions should have keys `feature_name`
        and values functions that take arguments (t, m, e); in the case of a
        dask graph, these arrays should be referenced as 't', 'm', 'e',
        respectively, and any values with keys present in `features_to_use`
        will be computed.
    scheduler : function, optional
        `dask` scheduler function used to perform feature extraction
        computation. Defaults to `dask.threaded.get`.
    raise_exceptions : bool, optional
        If True, exceptions during feature computation are raised immediately;
        if False, exceptions are supressed and `np.nan` is returned for the
        given feature and any dependent features. Defaults to True.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns containing feature values, indexed by name.
    """
    all_time_series = [delayed(time_series.load, pure=True)(ts_path)
                       for ts_path in ts_paths]
    all_features = [delayed(featurize_single_ts, pure=True)(ts, features_to_use,
                                                            custom_script_path,
                                                            custom_functions,
                                                            raise_exceptions)
                    for ts in all_time_series]
    names, meta_feats, all_labels = zip(*[(ts.name, ts.meta_features, ts.label)
                                          for ts in all_time_series])
    result = delayed(assemble_featureset, pure=True)(all_features,
                                                     meta_features_list=meta_feats,
                                                     names=names)
    fset, labels = dask.compute(result, all_labels, scheduler=scheduler)

    return fset, labels


def impute_featureset(fset, strategy='constant', value=None, max_value=1e20,
                      inplace=False):
    """Replace NaN/Inf values with imputed values as defined by `strategy`.
    Output should satisfy `sklearn.validation.assert_all_finite` so that
    training a model will not produce an error.

    Parameters
    ----------
    strategy : str, optional
    The imputation strategy. Defaults to 'constant'.

        - 'constant': replace all missing with `value`
        - 'mean': replace all missing with mean along `axis`
        - 'median': replace all missing with median along `axis`
        - 'most_frequent': replace all missing with mode along `axis`

    value : float or None, optional
        Replacement value to use for `strategy='constant'`. Defaults to
        `None`, in which case a very large negative value is used (a
        good choice for e.g. random forests).

    max_value : float, optional
        Maximum (absolute) value above which values are treated as infinite.
        Used to prevent overflow when fitting `sklearn` models.

    inplace : bool, optional
        If True, fill in place. If False, return a copy.

    Returns
    -------
    pd.DataFrame
        Feature data frame wth no missing/infinite values.
    """
    if not inplace:
        fset = fset.copy()
    fset.values[np.isnan(fset.values)] = np.inf  # avoid NaN comparison warnings
    fset.values[np.abs(fset.values) > max_value] = np.nan
    if strategy == 'constant':
        if value is None:
            # If no fill-in value is provided, use a large negative value
            value = -2. * np.nanmax(np.abs(fset.values))
        fset.fillna(value, inplace=True)
    elif strategy in ('mean', 'median', 'most_frequent'):
        imputer = Imputer(strategy=strategy)
        fset.values[:] = imputer.fit_transform(fset.values)
    else:
        raise NotImplementedError("Imputation strategy '{}' not"
                                  "recognized.".format(strategy))
    return fset


def save_featureset(fset, path, **kwargs):
    """Save feature DataFrame in .npz format.

    Can optionally store class labels/targets and other metadata. All other
    keyword arguments will be passed on to `np.savez`; data frames are saved as
    record arrays and converted back into data frames by `load_featureset`.

    Parameters
    ----------
    fset : pd.DataFrame
        Feature data frame to be saved.
    path : str
        Path to store feature data.
    kwargs : dict of array or data frame
        Additional keyword arguments, e.g.:
        labels -> class labels
        preds -> predicted class labels
        pred_probs -> (n_sample, n_class) data frame of class probabilities
    """
    # Transpose to properly handle MultiIndex columns
    kwargs['features'] = fset.T

    for k, v in kwargs.items():
        if isinstance(v, pd.DataFrame):
            arr = v.to_records()
            dt_list = arr.dtype.descr
            # Change type of indices from object to str
            for i, (name, dt) in enumerate(dt_list):
                if dt.endswith('O'):
                    size = max(len(x) for x in arr['index'])
                    dt_list[i] = (name, 'U' + str(size))
                dt_list[i] = (str(name),) + dt_list[i][1:]  # avoid Py2 unicode
            kwargs[k] = arr.astype(dt_list)

        # Ignore null values, e.g. for unlabeled data
        if all(el is None for el in v):
            kwargs[k] = []

    # Bypass savez to allow for `allow_pickle` keyword
    # See https://github.com/numpy/numpy/pull/5770
    np.lib.npyio._savez(path, [], kwargs, compress=True, allow_pickle=False)


def load_featureset(path):
    """Load feature DataFrame from .npz file.

    Feature information is returned as a single DataFrame, while any other
    arrays that were saved (class labels/predictions, etc.) are returned in a
    single dictionary.

    Parameters
    ----------
    path : str
        Path where feature data is stored.

    Returns
    -------
    pd.DataFrame
        Feature data frame to be saved.
    dict
        Additional variables passed to `save_featureset`, including labels, etc.
    """
    with np.load(path, allow_pickle=False) as npz_file:
        data = dict(npz_file)

    # Transpose to properly handle MultiIndex columns
    fset = pd.DataFrame.from_records(data.pop('features'),
                                     index=['feature', 'channel']).T
    features, channels = zip(*fset.columns)
    channels = [int(c) if str(c).isdigit() else '' for c in channels]
    fset.columns = pd.MultiIndex.from_tuples(list(zip(features, channels)),
                                             names=['feature', 'channel'])

    for k, v in data.items():
        if len(v.dtype) > 0:
            data[k] = pd.DataFrame.from_records(v, index='index')

    return fset, data



import numpy as np
import numpy.testing as npt

import time
import os
import shutil
import glob

from cesium.features.tests.util import (generate_features, irregular_random,
                                        regular_periodic, irregular_periodic)


def test_amplitude():
    """Test features related to amplitude/magnitude percentiles."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['amplitude'])
    npt.assert_allclose(f['amplitude'], (max(values) - min(values)) /2.)

    f = generate_features(times, values, errors, ['percent_amplitude'])
    max_scaled = 10**(-0.4 * max(values))
    min_scaled = 10**(-0.4 * min(values))
    med_scaled = 10**(-0.4 * np.median(values))
    peak_from_median = max(abs((max_scaled - med_scaled) / med_scaled),
            abs((min_scaled - med_scaled)) / med_scaled)
    npt.assert_allclose(f['percent_amplitude'], peak_from_median, rtol=5e-4)

    f = generate_features(times, values, errors, ['percent_difference_flux_percentile'])
    band_offset = 13.72
    w_m2 = 10**(-0.4*(values+band_offset)-3)  # 1 erg/s/cm^2 = 10^-3 w/m^2
    npt.assert_allclose(f['percent_difference_flux_percentile'], np.diff(
        np.percentile(w_m2, [5, 95])) / np.median(w_m2))

    f = generate_features(times, values, errors, ['flux_percentile_ratio_mid20'])
    npt.assert_allclose(f['flux_percentile_ratio_mid20'],
                            np.diff(np.percentile(w_m2, [40, 60])) /
                            np.diff(np.percentile(w_m2, [5, 95])))

    f = generate_features(times, values, errors, ['flux_percentile_ratio_mid35'])
    npt.assert_allclose(f['flux_percentile_ratio_mid35'],
                            np.diff(np.percentile(w_m2, [32.5, 67.5])) /
                            np.diff(np.percentile(w_m2, [5, 95])))

    f = generate_features(times, values, errors, ['flux_percentile_ratio_mid50'])
    npt.assert_allclose(f['flux_percentile_ratio_mid50'],
                            np.diff(np.percentile(w_m2, [25, 75])) /
                            np.diff(np.percentile(w_m2, [5, 95])))

    f = generate_features(times, values, errors, ['flux_percentile_ratio_mid65'])
    npt.assert_allclose(f['flux_percentile_ratio_mid65'],
                            np.diff(np.percentile(w_m2, [17.5, 82.5])) /
                            np.diff(np.percentile(w_m2, [5, 95])))

    f = generate_features(times, values, errors, ['flux_percentile_ratio_mid80'])
    npt.assert_allclose(f['flux_percentile_ratio_mid80'],
                            np.diff(np.percentile(w_m2, [10, 90])) /
                            np.diff(np.percentile(w_m2, [5, 95])))


# AR-IS features are currently ignored
"""
def test_ar_is():
    times = np.linspace(0, 500, 201)
    theta = 0.95
    sigma = 0.0
    values = theta ** (times/250.) + sigma*np.random.randn(len(times))
    errors = 1e-4*np.ones(len(times))

    f = generate_features(times, values, errors, ['ar_is_theta'])
    npt.assert_allclose(f['ar_is_theta'], theta, atol=3e-2)

    f = generate_features(times, values, errors, ['ar_is_sigma'])
    npt.assert_allclose(f['ar_is_sigma'], sigma, atol=1e-5)

# Hard-coded values from reference data set
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['ar_is_theta'])
    npt.assert_allclose(f['ar_is_theta'], 5.9608609865491405e-06, rtol=1e-3)
    f = generate_features(times, values, errors, ['ar_is_sigma'])
    npt.assert_allclose(f['ar_is_sigma'], 1.6427095072108497, rtol=1e-3)
"""


# The smoothed model fit in lcmodel seems insanely oversmoothed; in general
# all these extractors are a complete mess, and are currently ignored
"""
def test_lcmodel():
    frequencies = np.hstack((1., np.zeros(len(WAVE_FREQS)-1)))
    amplitudes = np.zeros((len(frequencies),4))
    amplitudes[0,:] = [2,0,0,0]
    phase = 0.0
    times, values, errors = regular_periodic(frequencies, amplitudes, phase)

    f = generate_features(times, values, errors, ['lcmodel'])
    lc_feats = f['lcmodel']

    # Median is zero for this data, so crossing median = changing sign
    incr_median_crosses = sum((np.abs(np.diff(np.sign(values))) > 1) &
                             (np.diff(values) > 0))
    npt.assert_allclose(lc_feats['median_n_per_day'], (incr_median_crosses+1) /
                        (max(times)-min(times)))
"""


def test_lomb_scargle_fast_regular():
    """Test gatspy's fast Lomb-Scargle period estimate on regularly-sampled
    periodic data.

    Note: this model fits only a single sinusoid with no additional harmonics,
    so we use only 1 frequency and 1 amplitude to generate test data.
    """
    frequencies = np.array([4])
    amplitudes = np.array([[1]])
    phase = 0.1
    times, values, errors = regular_periodic(frequencies, amplitudes, phase)
    f = generate_features(times, values, errors, ['period_fast'])

    npt.assert_allclose(f['period_fast'], 1. / frequencies[0], rtol=5e-4)


def test_lomb_scargle_fast_irregular():
    """Test gatspy's fast Lomb-Scargle period estimate on irregularly-sampled
    periodic data.

    Note: this model fits only a single sinusoid with no additional harmonics,
    so we use only 1 frequency and 1 amplitude to generate test data.
    """
    frequencies = np.array([4])
    amplitudes = np.array([[1]])
    phase = 0.1
    times, values, errors = irregular_periodic(frequencies, amplitudes, phase)
    f = generate_features(times, values, errors, ['period_fast'])

    npt.assert_allclose(f['period_fast'], 1. / frequencies[0], rtol=3e-2)


def test_max():
    """Test maximum value feature."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['maximum'])
    npt.assert_equal(f['maximum'], max(values))


def test_max_slope():
    """Test maximum slope feature, which finds the INDEX of the largest slope."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['max_slope'])
    slopes = np.diff(values) / np.diff(times)
    npt.assert_allclose(f['max_slope'], np.max(np.abs(slopes)))


def test_median_absolute_deviation():
    """Test median absolute deviation (from the median) feature."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['median_absolute_deviation'])
    npt.assert_allclose(f['median_absolute_deviation'], np.median(np.abs(values -
        np.median(values))))


def test_percent_close_to_median():
    """Test feature which finds the percentage of points near the median value."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors,
            ['percent_close_to_median'])
    amplitude = (max(values) - min(values)) / 2.
    within_buffer = np.abs(values - np.median(values)) < 0.2*amplitude
    npt.assert_allclose(f['percent_close_to_median'], np.mean(within_buffer))


def test_median():
    """Test median value feature."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['median'])
    npt.assert_allclose(f['median'], np.median(values))


def test_min():
    """Test minimum value feature."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['minimum'])
    npt.assert_equal(f['minimum'], min(values))


# These features are currently ignored
"""
def test_phase_dispersion():
# Frequency chosen to lie on the relevant search grid
    frequencies = np.hstack((5.36, np.zeros(len(WAVE_FREQS)-1)))
    amplitudes = np.zeros((len(frequencies),4))
    amplitudes[0,:] = [1,0,0,0]
    phase = 0.1
    times, values, errors = regular_periodic(frequencies, amplitudes, phase)
    f = generate_features(times, values, errors, ['phase_dispersion_freq0'])
    npt.assert_allclose(f['phase_dispersion_freq0'], frequencies[0])

    f = generate_features(times, values, errors, ['freq1_freq'])
    lomb_freq = f['freq1_freq']
    f = generate_features(times, values, errors, ['ratio_PDM_LS_freq0'])
    npt.assert_allclose(f['ratio_PDM_LS_freq0'], frequencies[0]/lomb_freq)
"""


def test_qso_features():
    """Test features which measure fit of QSO model.

    Reference values are hard-coded values from previous implementation; not sure
    of examples with a closed-form solution.
    """
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors,
            ['qso_log_chi2_qsonu', 'qso_log_chi2nuNULL_chi2nu'])
    npt.assert_allclose(f['qso_log_chi2_qsonu'], 6.9844064754)
    npt.assert_allclose(f['qso_log_chi2nuNULL_chi2nu'], -0.456526327522)


def test_skew():
    """Test statistical skew feature."""
    from scipy import stats
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['skew'])
    npt.assert_allclose(f['skew'], stats.skew(values))


def test_std():
    """Test standard deviation feature."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['std'])
    npt.assert_allclose(f['std'], np.std(values))


def test_stetson():
    """Test Stetson variability features."""
    times, values, errors = irregular_random(size=201)
    f = generate_features(times, values, errors, ['stetson_j'])
    # Stetson mean approximately equal to standard mean for large inputs
    dists = np.sqrt(float(len(values)) / (len(values) - 1.)) * (values - np.mean(values)) / 0.1
    npt.assert_allclose(f['stetson_j'],
                        np.mean(np.sign(dists**2-1)*np.sqrt(np.abs(dists**2-1))),
                        rtol=1e-2)
    # Stetson_j should be somewhat close to (scaled) variance for normal data
    npt.assert_allclose(f['stetson_j']*0.1, np.var(values), rtol=2e-1)
    # Hard-coded original value
    npt.assert_allclose(f['stetson_j'], 7.591347175195703)

    f = generate_features(times, values, errors, ['stetson_k'])
    npt.assert_allclose(f['stetson_k'], 1./0.798 * np.mean(np.abs(dists)) / np.sqrt(np.mean(dists**2)), rtol=5e-4)
    # Hard-coded original value
    npt.assert_allclose(f['stetson_k'], 1.0087218792719013)


def test_weighted_average():
    """Test weighted average and distance from weighted average features."""
    times, values, errors = irregular_random()
    f = generate_features(times, values, errors, ['weighted_average'])
    weighted_avg = np.average(values, weights=1. / (errors**2))
    weighted_var = np.average((values - weighted_avg)**2,
                              weights=1. / (errors**2))
    npt.assert_allclose(f['weighted_average'], weighted_avg)

    dists_from_weighted_avg = values - weighted_avg
    stds_from_weighted_avg = (dists_from_weighted_avg /
            np.sqrt(weighted_var))

    f = generate_features(times, values, errors,
                                      ['percent_beyond_1_std'])
    npt.assert_equal(f['percent_beyond_1_std'],
                     np.mean(np.abs(stds_from_weighted_avg) > 1.))



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('cesium', parent_package, top_path)
    config.add_subpackage('features')
    config.add_data_dir('tests')
    config.add_data_dir('tests/data')
    config.add_data_dir('features/tests/data')
    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup

    config = configuration(top_path='').todict()
    setup(**config)



"""
=================================
Epilepsy Detection Using EEG Data
=================================

In this example we'll use the |cesium|_ library to compare
various techniques for epilepsy detection using a classic EEG time series dataset from
`Andrzejak et al.  <http://www.meb.uni-bonn.de/epileptologie/science/physik/eegdata.html>`_.
The raw data are separated into five classes: Z, O, N, F, and S; we will consider a
three-class classification problem of distinguishing normal (Z, O), interictal (N, F), and
ictal (S) signals.

The overall workflow consists of three steps: first, we "featurize" the time series by
selecting some set of mathematical functions to apply to each; next, we build some
classification models which use these features to distinguish between classes;
finally, we validate our models by generating predictions for some unseen
holdout set and comparing them to the true class labels.

First, we'll load the data and inspect a representative time series from each class:

.. |cesium| replace:: ``cesium``
.. _cesium: https://github.com/cesium-ml/cesium
"""

import numpy as np
import matplotlib.pyplot as plt
import seaborn; seaborn.set()

from cesium import datasets

eeg = datasets.fetch_andrzejak()

# Group together classes (Z, O), (N, F), (S) as normal, interictal, ictal
eeg["classes"] = eeg["classes"].astype("U16") #  allocate memory for longer class names
eeg["classes"][np.logical_or(eeg["classes"]=="Z", eeg["classes"]=="O")] = "Normal"
eeg["classes"][np.logical_or(eeg["classes"]=="N", eeg["classes"]=="F")] = "Interictal"
eeg["classes"][eeg["classes"]=="S"] = "Ictal"

fig, ax = plt.subplots(1, len(np.unique(eeg["classes"])), sharey=True)
for label, subplot in zip(np.unique(eeg["classes"]), ax):
    i = np.where(eeg["classes"] == label)[0][0]
    subplot.plot(eeg["times"][i], eeg["measurements"][i])
    subplot.set(xlabel="time (s)", ylabel="signal", title=label)


###############################################################################
# Featurization
# -------------
# Once the data is loaded, we can generate features for each time series using the
# |cesium.featurize|_ module. The ``featurize`` module includes many built-in
# choices of features which can be applied for any type of time series data;
# here we've chosen a few generic features that do not have any special
# biological significance.
#
# By default, the time series will featurized in parallel using the
# ``dask.threaded`` scheduler; other approaches, including serial and
# distributed approaches, can be implemented by passing in other ``dask``
# schedulers as the ``get`` argument to ``featurize_time_series``.
#
# .. |cesium.featurize| replace:: ``cesium.featurize``
# .. _cesium.featurize: http://cesium-ml.org/docs/api/cesium.featurize.html

from cesium import featurize
features_to_use = ["amplitude",
                   "percent_beyond_1_std",
                   "maximum",
                   "max_slope",
                   "median",
                   "median_absolute_deviation",
                   "percent_close_to_median",
                   "minimum",
                   "skew",
                   "std",
                   "weighted_average"]
fset_cesium = featurize.featurize_time_series(times=eeg["times"],
                                              values=eeg["measurements"],
                                              errors=None,
                                              features_to_use=features_to_use)
print(fset_cesium.head())

###############################################################################
# The output of ``featurize_time_series`` is a ``pandas.DataFrame`` which contains all
# the feature information needed to train a machine learning model: feature
# names are stored as column indices (as well as channel numbers, as we'll see
# later for multi-channel data), and the time series index/class label are
# stored as row indices.

###############################################################################
# Custom feature functions
# ~~~~~~~~~~~~~~~~~~~~~~~~
# Custom feature functions not built into ``cesium`` may be passed in using the
# ``custom_functions`` keyword, either as a dictionary ``{feature_name: function}``, or as a
# `dask graph <http://dask.pydata.org/en/latest/custom-graphs.html>`_. Functions should take
# three arrays ``times, measurements, errors`` as inputs; details can be found in the
# ``cesium.featurize``
# `documentation <http://cesium-ml.org/docs/api/cesium.featurize.html>`_.
# Here we'll compute five standard features for EEG analysis provided by
# `Guo et al. (2012) <http://linkinghub.elsevier.com/retrieve/pii/S0957417411003253)>`_:

import numpy as np
import scipy.stats

def mean_signal(t, m, e):
    return np.mean(m)

def std_signal(t, m, e):
    return np.std(m)

def mean_square_signal(t, m, e):
    return np.mean(m ** 2)

def abs_diffs_signal(t, m, e):
    return np.sum(np.abs(np.diff(m)))

def skew_signal(t, m, e):
    return scipy.stats.skew(m)

###############################################################################
# Now we'll pass the desired feature functions as a dictionary via the
# ``custom_functions`` keyword argument.

guo_features = {
    "mean": mean_signal,
    "std": std_signal,
    "mean2": mean_square_signal,
    "abs_diffs": abs_diffs_signal,
    "skew": skew_signal
}

fset_guo = featurize.featurize_time_series(times=eeg["times"], values=eeg["measurements"],
                                           errors=None,
                                           features_to_use=list(guo_features.keys()),
                                           custom_functions=guo_features)
print(fset_guo.head())

###############################################################################
# Multi-channel time series
# ~~~~~~~~~~~~~~~~~~~~~~~~~
# The EEG time series considered here consist of univariate signal measurements along a
# uniform time grid. But ``featurize_time_series`` also accepts multi-channel
# data; to demonstrate this, we will decompose each signal into five frequency
# bands using a discrete wavelet transform as suggested by
# `Subasi (2005) <http://www.sciencedirect.com/science/article/pii/S0957417404001745>`_,
# and then featurize each band separately using the five functions from above.

import pywt

n_channels = 5
eeg["dwts"] = [pywt.wavedec(m, pywt.Wavelet("db1"), level=n_channels-1)
               for m in eeg["measurements"]]
fset_dwt = featurize.featurize_time_series(times=None, values=eeg["dwts"], errors=None,
                                           features_to_use=list(guo_features.keys()),
                                           custom_functions=guo_features)
print(fset_dwt.head())

###############################################################################
# The output featureset has the same form as before, except now the ``channel``
# component of the column index is used to index the features by the
# corresponding frequency band.

###############################################################################
# Model Building
# --------------
# Featuresets produced by ``cesium.featurize`` are compatible with the ``scikit-learn``
# API. For this example, we'll test a random forest classifier for the
# built-in ``cesium`` features, and a 3-nearest neighbors classifier for the
# others, as suggested by
# `Guo et al. (2012) <http://linkinghub.elsevier.com/retrieve/pii/S0957417411003253>`_.

from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.model_selection import train_test_split

train, test = train_test_split(np.arange(len(eeg["classes"])), random_state=0)

model_cesium = RandomForestClassifier(n_estimators=128, max_features="auto",
                                      random_state=0)
model_cesium.fit(fset_cesium.iloc[train], eeg["classes"][train])

model_guo = KNeighborsClassifier(3)
model_guo.fit(fset_guo.iloc[train], eeg["classes"][train])

model_dwt = KNeighborsClassifier(3)
model_dwt.fit(fset_dwt.iloc[train], eeg["classes"][train])

###############################################################################
# Prediction
# ----------
# Making predictions for new time series based on these models follows the same
# pattern: first the time series are featurized using ``featurize_time_series``,
# and then predictions are made based on these features using the ``predict``
# method of the ``scikit-learn`` model.

from sklearn.metrics import accuracy_score

preds_cesium = model_cesium.predict(fset_cesium)
preds_guo = model_guo.predict(fset_guo)
preds_dwt = model_dwt.predict(fset_dwt)

print("Built-in cesium features: training accuracy={:.2%}, test accuracy={:.2%}".format(
          accuracy_score(preds_cesium[train], eeg["classes"][train]),
          accuracy_score(preds_cesium[test], eeg["classes"][test])))
print("Guo et al. features: training accuracy={:.2%}, test accuracy={:.2%}".format(
          accuracy_score(preds_guo[train], eeg["classes"][train]),
          accuracy_score(preds_guo[test], eeg["classes"][test])))
print("Wavelet transform features: training accuracy={:.2%}, test accuracy={:.2%}".format(
          accuracy_score(preds_dwt[train], eeg["classes"][train]),
          accuracy_score(preds_dwt[test], eeg["classes"][test])))

###############################################################################
# The workflow presented here is intentionally simplistic and omits many important steps
# such as feature selection, model parameter selection, etc., which may all be
# incorporated just as they would for any other ``scikit-learn`` analysis.
# But with essentially three function calls (``featurize_time_series``,
# ``model.fit``, and ``model.predict``), we are able to build a
# model from a set of time series and make predictions on new, unlabeled data. In
# upcoming posts we'll introduce the web frontend for ``cesium`` and describe how
# the same analysis can be performed in a browser with no setup or coding required.



import numpy as np
from scipy import stats


def max_slope(t, x):
    """Compute the largest rate of change in the observed data."""
    slopes = np.diff(x) / np.diff(t)
    return np.max(np.abs(slopes))


def maximum(x):
    """Maximum observed value."""
    return np.max(x)


def median(x):
    """Median of observed values."""
    return np.median(x)


def median_absolute_deviation(x):
    """Median absolute deviation (from the median) of the observed values."""
    return np.median(np.abs(x - np.median(x)))


def minimum(x):
    """Minimum observed value."""
    return np.min(x)


def percent_beyond_1_std(x, e):
    """Percentage of values more than 1 std. dev. from the weighted average."""
    dists_from_mu = x - weighted_average(x, e)
    return np.mean(np.abs(dists_from_mu) > weighted_std_dev(x, e))


def percent_close_to_median(x, window_frac=0.1):
    """Percentage of values within window_frac*(max(x)-min(x)) of median."""
    window = (x.max() - x.min()) * window_frac
    return np.mean(np.abs(x - np.median(x)) < window)


def skew(x):
    """Skewness of a dataset. Approximately 0 for Gaussian data."""
    return stats.skew(x)


def std(x):
    """Standard deviation of observed values."""
    return np.std(x)


def weighted_average(x, e):
    """Arithmetic mean of observed values, weighted by measurement errors."""
    return np.average(x, weights=1. / (e**2))


def weighted_average_std_err(x, e):
    """
    Standard deviation of the sample weighted average of values x with
    measurement errors e.

    Note: this is not the same as the weighted sample standard deviation;
    this value only quantifies the measurement errors, not the dispersion of
    the data.
    """
    return np.sqrt(1.0 / np.sum(e**2))


def weighted_std_dev(x, e):
    """Standard deviation of observed values, weighted by measurement errors."""
    return np.sqrt(np.average((x - weighted_average(x, e))**2,
                              weights=1. / (e**2)))



import copy
from collections import Iterable
import numpy as np


__all__ = ['load', 'TimeSeries', 'DEFAULT_MAX_TIME',
           'DEFAULT_ERROR_VALUE']


DEFAULT_MAX_TIME = 1.0
DEFAULT_ERROR_VALUE = 1e-4


def _ndim(x):
    """Return number of dimensions for a (possibly ragged) array."""
    n = 0
    while isinstance(x, Iterable):
        x = x[0]
        n += 1
    return n


def _compatible_shapes(x, y):
    """Check recursively that iterables x and y (and each iterable contained
    within, if applicable), have compatible sizes.
    """
    if hasattr(x, 'shape') and hasattr(y, 'shape'):
        return x.shape == y.shape
    else:
        return (len(x) == len(y) and all(np.shape(x_i) == np.shape(y_i)
                                         for x_i, y_i in zip(x, y)))


def _default_values_like(old_values, value=None, upper=None):
    """Creates a range of default values with the same shape as the input
    `old_values`. If `value` is provided then each entry will equal `value`;
    if `upper` is provided then the values will be linearly-spaced from 0 to
    `upper`.

    Parameters
    ----------
    old_values : (n,) or (p,n) array or list of (n,) arrays
        Input array(s), typically time series measurements for which default
        time or error values need to be inferred.
    value : float, optional
        Value that each output entry will be set to (omitted if `upper` is
        provided).
    upper : float, optional
        Upper bound of range of linearly-spaced output entries (omitted if
        `value` is provided).
    """
    if value and upper:
        raise ValueError("Only one of `value` or `upper` may be proivded.")
    elif value is not None:
        lower = value
        upper = value
    elif upper is not None:
        lower = 0.
    else:
        raise ValueError("Either `value` or `upper` must be provided.")

    new_values = copy.deepcopy(old_values)
    if _ndim(old_values) == 1 or (isinstance(old_values, np.ndarray) and 1 in
                                  old_values.shape):
        new_values[:] = np.linspace(lower, upper, len(new_values))
    else:
        for new_array in new_values:
            new_array[:] = np.linspace(lower, upper, len(new_array))

    return new_values


def _make_array_if_possible(x):
    """Helper function to cast (1, n) arrays to (n,) arrrays, or uniform lists
    of arrays to (p, n) arrays.
    """
    try:
        x = np.asfarray(x).squeeze()
    except ValueError:
        pass
    return x


def load(ts_path):
    """Load serialized TimeSeries from .npz file."""
    with np.load(ts_path) as npz_file:
        data = dict(npz_file)

    for key in ['time', 'measurement', 'error']:
        if key not in data:  # combine channel arrays into list
            n_channels = sum(1 for c in data.keys() if key in c)  # time0, ...
            data[key] = [data[key + str(i)] for i in range(n_channels)]

    # Convert 0d arrays to single values
    if 'name' in data:
        data['name'] = data['name'].item()
    if 'label' in data:
        data['label'] = data['label'].item()

    return TimeSeries(t=data.get('time'), m=data.get('measurement'),
                      e=data.get('error'),
                      meta_features=dict(zip(data['meta_feat_names'],
                                             data['meta_feat_values'])),
                      name=data.get('name'), label=data.get('label'))


class TimeSeries(object):
    """Class representing a single time series of measurements and metadata.

    A `TimeSeries` object encapsulates a single set of time-domain
    measurements, along with any metadata describing the observation.
    Typically the observations will consist of times, measurements, and
    (optionally) measurement errors. The measurements can be scalar- or
    vector-valued (i.e., "multichannel"); for multichannel measurements, the
    times and errors can also be vector-valued, or they can be shared across
    all channels of measurement.

    Attributes
    ----------
    time : (n,) or (p, n) array or list of (n,) arrays
        Array(s) of times corresponding to measurement values. If `measurement`
        is two-dimensional, this can be one-dimensional (same times for each
        channel) or two-dimensional (different times for each channel). If
        `time` is one-dimensional then it will be broadcast to match
        `measurement.shape`.
    measurement : (n,) or (p, n) array or list of (n,) arrays
        Array(s) of measurement values; can be two-dimensional for
        multichannel data. In the case of multichannel data with different
        numbers of measurements for each channel, `measurement` will be a list
        of arrays instead of a single two-dimensional array.
    error : (n,) or (p, n) array or list of (n,) arrays
        Array(s) of measurement errors for each value. If `measurement` is
        two-dimensional, this can be one-dimensional (same times for each
        channel) or two-dimensional (different times for each channel).
        If `error` is one-dimensional then it will be broadcast match
        `measurement.shape`.
    label : str, float, or None
        Class label or regression target for the given time series (if
        applicable).
    meta_features : dict
        Dictionary of feature names/values specified independently of the
        featurization process in `featurize`.
    name : str or None
        Identifying name for the given time series (if applicable).
        Typically the name of the raw data file from which the time series was
        created.
    path : str or None
        Path to the file where the time series is stored on disk (if
        applicable).
    channel_names : list of str
        List of names of channels of measurement; by default these are simply
        `channel_{i}`, but can be arbitrary depending on the nature of the
        different measurement channels.
    """
    def __init__(self, t=None, m=None, e=None, label=None, meta_features={},
                 name=None, path=None, channel_names=None):
        """Create a `TimeSeries` object from measurement values/metadata.

        See `TimeSeries` documentation for parameter values.
        """
        if t is None and m is None:
            raise ValueError("Either times or measurements must be provided.")
        elif m is None:
            m = _default_values_like(t, value=np.nan)

        # If m is 1-dimensional, so are t and e
        if _ndim(m) == 1:
            self.n_channels = 1
            if t is None:
                t = _default_values_like(m, upper=DEFAULT_MAX_TIME)
            if e is None:
                e = _default_values_like(m, value=DEFAULT_ERROR_VALUE)
        # If m is 2-dimensional, t and e could be 1d or 2d; default is 1d
        elif isinstance(m, np.ndarray) and m.ndim == 2:
            self.n_channels = len(m)
            if t is None:
                t = _default_values_like(m[0], upper=DEFAULT_MAX_TIME)
            if e is None:
                e = _default_values_like(m[0], value=DEFAULT_ERROR_VALUE)
        # If m is ragged (list of 1d arrays), t and e should also be ragged
        elif _ndim(m) == 2:
            self.n_channels = len(m)
            if t is None:
                t = _default_values_like(m, upper=DEFAULT_MAX_TIME)
            if e is None:
                e = _default_values_like(m, value=DEFAULT_ERROR_VALUE)
        else:
            raise ValueError("m must be a 1D or 2D array, or a 2D list of"
                             " arrays.")

        self.time = _make_array_if_possible(t)
        self.measurement = _make_array_if_possible(m)
        self.error = _make_array_if_possible(e)
        self.sort()  # re-order by time before broadcasting

        if _ndim(self.time) == 1 and _ndim(self.measurement) == 2:
            if isinstance(self.measurement, np.ndarray):
                self.time = np.broadcast_to(self.time, self.measurement.shape)
            else:
                raise ValueError("Times for each channel must be provided if m"
                                 " is a ragged array.")

        if _ndim(self.error) == 1 and _ndim(self.measurement) == 2:
            if isinstance(self.measurement, np.ndarray):
                self.error = np.broadcast_to(self.error, self.measurement.shape)
            else:
                raise ValueError("Errors for each channel must be provided if"
                                 " m is a ragged array.")

        if not (_compatible_shapes(self.measurement, self.time) and
                _compatible_shapes(self.measurement, self.error)):
            raise ValueError("times, values, errors are not of compatible"
                             " types/sizes. Please refer to the docstring"
                             " for list of allowed input types.")

        self.label = label
        self.meta_features = dict(meta_features)
        self.name = name
        self.path = path
        if channel_names is None:
            self.channel_names = ["channel_{}".format(i)
                                  for i in range(self.n_channels)]
        else:
            self.channel_names = channel_names

    def channels(self):
        """Iterates over measurement channels (whether one or multiple)."""
        t_channels = self.time
        m_channels = self.measurement
        e_channels = self.error
        if isinstance(self.time, np.ndarray) and self.time.ndim == 1:
            t_channels = np.broadcast_to(self.time,
                                         (self.n_channels, len(self.time)))
        if (isinstance(self.measurement, np.ndarray) and
                self.measurement.ndim == 1):
            m_channels = np.broadcast_to(self.measurement,
                                         (self.n_channels, len(self.measurement)))
        if isinstance(self.error, np.ndarray) and self.error.ndim == 1:
            e_channels = np.broadcast_to(self.error,
                                         (self.n_channels, len(self.error)))
        return zip(t_channels, m_channels, e_channels)

    def sort(self):
        """Sort times, measurements, and errors by time."""
        if _ndim(self.time) == 1:
            inds = np.argsort(self.time)
            self.time = self.time[inds]
            if _ndim(self.measurement) == 1:
                self.measurement = self.measurement[inds]
            else:
                for i in range(len(self.measurement)):
                   self.measurement[i] = self.measurement[i][inds]
            if _ndim(self.error) == 1:
                self.error = self.error[inds]
            else:
                for i in range(len(self.error)):
                   self.error[i] = self.error[i][inds]
        else:  # if time is 2d, so are measurement and error
            for i in range(len(self.time)):
                inds = np.argsort(self.time[i])
                self.time[i] = self.time[i][inds]
                self.measurement[i] = self.measurement[i][inds]
                self.error[i] = self.error[i][inds]


    def save(self, path=None):
        """Store TimeSeries object as a single .npz file.

        Attributes are stored in the following arrays:
            - time
            - measurement
            - error
            - meta_feat_names
            - meta_feat_values
            - name
            - label

        If `path` is omitted then the `path` attribute from the TimeSeries
        object is used.
        """
        if path is None:
            path = self.path

        data = {'meta_feat_names': list(self.meta_features.keys()),
                'meta_feat_values': list(self.meta_features.values())}

        for key in ['time', 'measurement', 'error']:
            value = getattr(self, key)
            if isinstance(value, np.ndarray):
                data[key] = value
            else:  # list of arrays -> save each channel separately
                for i, value_i in enumerate(value):
                    data[key + str(i)] = value_i

        if self.name:
            data['name'] = self.name
        if self.label:
            data['label'] = self.label
        np.savez(path, **data)



import os
from cesium import util
import numpy.testing as npt


def test_shorten_fname():
    """Test util.shorten_fname"""
    npt.assert_equal(util.shorten_fname("/a/b/c/name.ext"), "name")
    npt.assert_equal(util.shorten_fname("path/to/fname.name2.ext2"),
                     "fname.name2")
    npt.assert_equal(util.shorten_fname("fname.ext"), "fname")
    npt.assert_equal(util.shorten_fname("fname"), "fname")


def test_remove_files(tmpdir):
    """Test util.remove_files"""
    # Pass in single path (non-list)
    fpath = os.path.join(str(tmpdir), "cesium.test")
    f = open(fpath, "w").close()
    assert os.path.exists(fpath)
    util.remove_files(fpath)
    assert not os.path.exists(fpath)

    # Pass in list of paths
    f = open(fpath, "w").close()
    assert os.path.exists(fpath)
    util.remove_files([fpath])
    assert not os.path.exists(fpath)

    # File does not exist, should not raise exception
    util.remove_files(fpath)



import numpy as np
import scipy.stats as stats
from ._lomb_scargle import lomb_scargle


def lomb_scargle_model(time, signal, error, sys_err=0.05, nharm=8, nfreq=3, tone_control=5.0):
    """Simultaneous fit of a sum of sinusoids by weighted least squares:
           y(t) = Sum_k Ck*t^k + Sum_i Sum_j A_ij sin(2*pi*j*fi*(t-t0)+phi_j),
           i=[1,nfreq], j=[1,nharm]

    Parameters
    ----------
    time : array_like
        Array containing time values.

    signal : array_like
        Array containing data values.

    error : array_like
        Array containing measurement error values.

    nharm : int
        Number of harmonics to fit for each frequency.

    nfreq : int
        Number of frequencies to fit.

    Returns
    -------
    dict
        Dictionary containing fitted parameter values. Parameters specific to
        a specific fitted frequency are stored in a list of dicts at
        model_dict['freq_fits'], each of which contains the output of
        fit_lomb_scargle(...)

    """

    dy0 = np.sqrt(error**2 + sys_err**2)

    wt = 1. / dy0**2
    time = time.copy() - min(time) # speeds up lomb_scargle code to have min(time)==0
    signal = signal.copy()

    chi0 = np.dot(signal**2, wt)

# TODO parametrize?
    f0 = 1. / max(time)
    df = 0.8 / max(time) # 20120202 :    0.1/Xmax
    fmax = 33. #pre 20120126: 10. # 25
    numf = int((fmax - f0) / df) # TODO !!! this is off by 1 point, fix?

    model_dict = {'freq_fits' : []}
    lambda0_range = [-np.log10(len(time)), 8] # these numbers "fix" the strange-amplitude effect
    for i in range(nfreq):
        if i == 0:
            fit = fit_lomb_scargle(time, signal, dy0, f0, df, numf,
                    tone_control=tone_control, lambda0_range=lambda0_range,
                    nharm=nharm, detrend_order=1)
            model_dict['trend'] = fit['trend_coef'][1]
        else:
            fit = fit_lomb_scargle(time, signal, dy0, f0, df, numf,
                    tone_control=tone_control, lambda0_range=lambda0_range,
                    nharm=nharm, detrend_order=0)
        model_dict['freq_fits'].append(fit)
        signal -= fit['model']
        model_dict['freq_fits'][-1]['resid'] = signal.copy()
        if i == 0:
            model_dict['varrat'] = np.dot(signal**2, wt) / chi0

    model_dict['nfreq'] = nfreq
    model_dict['nharm'] = nharm
    model_dict['chi2'] = fit['chi2']
    model_dict['f0'] = f0
    model_dict['df'] = df
    model_dict['numf'] = numf

    return model_dict


def lprob2sigma(lprob):
    """Translate a log_e(probability) to units of Gaussian sigmas."""
    if lprob > -36.:
        sigma = stats.norm.ppf(1. - 0.5 * np.exp(lprob))
    else:
        sigma = np.sqrt(np.log(2. / np.pi) - 2. * np.log(8.2) - 2. * lprob)
        f = 0.5 * np.log(2. / np.pi) - 0.5 * sigma**2 - np.log(sigma) - lprob
        sigma += f / (sigma + 1. / sigma)
    return sigma


def fit_lomb_scargle(time, signal, error, f0, df, numf, nharm=8, psdmin=6., detrend_order=0,
         freq_zoom=10., tone_control=5., lambda0=1., lambda0_range=[-8,6]):
    """Calls C implementation of Lomb Scargle sinusoid fitting, which fits a
    single frequency with nharm harmonics to the data. Called repeatedly by
    lomb_scargle_model in order to produce a fit with multiple distinct
    frequencies.

    Inputs:
    time : array_like
        Array containing time values.

    signal : array_like
        Array containing data values.

    error : array_like
        Array containing measurement error values.

    f0 : float
        Smallest frequency value to consider.

    df : float
        Step size for frequency grid search.

    numf : int
        Number of frequencies for frequency grid search.

    nharm : int
        Number of harmonics to fit.

    detrend_order : int
        Order of polynomial detrending.

    psdmin : int
        Refine periodogram values with larger psd using multi-harmonic fit

    nharm : int
        Number of harmonics to use in refinement

    lambda0 : float
        Typical value for regularization parameter

    lambda0_range : [float, float]
        Allowable range for log10 of regularization parameter

    Returns
    -------
    dict
        Dictionary describing various parameters of the multiharmonic fit at
        the best-fit frequency
    """
    ntime = len(time)

# For some reason we round this to the nearest even integer
    freq_zoom = round(freq_zoom/2.)*2.

# Polynomial terms
    coef = np.zeros(detrend_order + 1, dtype='float64')
    norm = np.zeros(detrend_order + 1, dtype='float64')

    wth0 = 1. / error
    s0 = np.dot(wth0, wth0)
    wth0 /= np.sqrt(s0)

    cn = signal * wth0
    coef[0] = np.dot(cn,wth0)
    cn0 = coef[0]
    norm[0] = 1.
    cn -= coef[0] * wth0
    vcn = 1.

    # np.sin's and cosin's for later
    tt = 2. * np.pi * time
    sinx,cosx = np.sin(tt*f0)*wth0,np.cos(tt*f0)*wth0
    sinx_step,cosx_step = np.sin(tt*df),np.cos(tt*df)
    sinx_back,cosx_back = -np.sin(tt*df/2.),np.cos(tt*df/2)
    sinx_smallstep,cosx_smallstep = np.sin(tt*df/freq_zoom),np.cos(tt*df/freq_zoom)

    npar = 2*nharm
    hat_matr = np.zeros((npar,ntime),dtype='float64')
    hat0 = np.zeros((npar,detrend_order+1),dtype='float64')
    hat_hat = np.zeros((npar,npar),dtype='float64')
    soln = np.zeros(npar,dtype='float64')
    psd = np.zeros(numf,dtype='float64')

    # Detrend the data and create the orthogonal detrending basis
    if detrend_order > 0:
        wth = np.zeros((detrend_order + 1, ntime),dtype='float64')
        wth[0,:] = wth0
    else:
        wth = wth0

    for i in range(detrend_order):
        f = wth[i,:] * tt / (2 * np.pi)
        for j in range(i+1):
            f -= np.dot(f, wth[j,:]) * wth[j,:]
        norm[i+1] = np.sqrt(np.dot(f,f))
        f /= norm[i+1]
        coef[i+1] = np.dot(cn,f)
        cn -= coef[i+1]*f
        wth[i+1,:] = f
        vcn += (f/wth0)**2

    chi0 = np.dot(cn,cn)
    varcn = chi0/(ntime-1-detrend_order)
    psdmin *= 2*varcn

    Tr = np.array(0., dtype='float64')
    ifreq = np.array(0, dtype='int32')
    lambda0 = np.array(lambda0 / s0, dtype='float64')
    lambda0_range = 10**np.array(lambda0_range, dtype='float64') / s0

    lomb_scargle(ntime, numf, nharm, detrend_order, psd, cn, wth, sinx, cosx,
            sinx_step, cosx_step, sinx_back, cosx_back, sinx_smallstep,
            cosx_smallstep, hat_matr, hat_hat, hat0, soln, chi0, freq_zoom,
            psdmin, tone_control, lambda0, lambda0_range, Tr, ifreq)

    hat_hat /= s0
    ii = np.arange(nharm, dtype='int32')
    soln[0:nharm] /= (1. + ii)**2
    soln[nharm:] /= (1. + ii)**2
    hat_matr0 = np.outer(hat0[:,0], wth0)
    for i in range(detrend_order):
        hat_matr0 += np.outer(hat0[:,i+1], wth[i+1,:])

    modl = np.dot(hat_matr.T, soln)
    coef0 = np.dot(soln, hat0)
    coef -= coef0
    hat_matr -= hat_matr0

    out_dict = {}
    out_dict['psd'] = psd
    out_dict['chi0'] = chi0 * s0
    if detrend_order > 0:
        out_dict['trend'] = np.dot(coef,wth)/wth0
    else:
        out_dict['trend'] = coef[0] + 0*wth0
    out_dict['model'] = modl/wth0 + out_dict['trend']

    j = psd.argmax()
    freq = f0 + df * j + (ifreq / freq_zoom - 1/2.) * df
    tt = (time * freq) % 1.
    out_dict['freq'] = freq
    out_dict['s0'] = s0
    out_dict['chi2'] = (chi0 - psd[j]) * s0
    out_dict['psd'] = psd[j] * 0.5 / varcn
    out_dict['lambda'] = lambda0 * s0
#    out_dict['gcv_weight'] = (1 - 3. / ntime) / Tr
    out_dict['trace'] = Tr
    out_dict['nu0'] = ntime - npar
    npars = (1 - Tr) * ntime / 2.
    out_dict['nu'] = ntime - npars
    out_dict['npars'] = npars

    A0, B0 = soln[0:nharm], soln[nharm:]
    hat_hat /= np.outer(np.hstack(((1.+ii)**2, (1.+ii)**2)), np.hstack(((1.+ii)**2, (1.+ii)**2)))
    err2 = np.diag(hat_hat)
    vA0, vB0 = err2[0:nharm], err2[nharm:]
    covA0B0 = hat_hat[(ii,nharm+ii)]

    vmodl = vcn/s0 + np.dot((hat_matr/wth0).T, np.dot(hat_hat, hat_matr/wth0))
    vmodl0 = vcn/s0 + np.dot((hat_matr0/wth0).T, np.dot(hat_hat, hat_matr0/wth0))
    out_dict['model_error'] = np.sqrt(np.diag(vmodl))
    out_dict['trend_error'] = np.sqrt(np.diag(vmodl0))

    amp = np.sqrt(A0**2 + B0**2)
    damp = np.sqrt(A0**2 * vA0 + B0**2 * vB0 + 2. * A0 * B0 * covA0B0) / amp
    phase = np.arctan2(B0, A0)
    rel_phase = phase - phase[0]*(1.+ii)
    rel_phase = np.arctan2(np.sin(rel_phase), np.cos(rel_phase))
    dphase = 0.*rel_phase
    for i in range(nharm - 1):
        j = i + 1
        v = np.array([-A0[0] * (1. + j) / amp[0]**2, B0[0] * (1. + j) / amp[0]**2, A0[j] / amp[j]**2, -B0[j] / amp[j]**2])
        jj = np.array([0, nharm, j, j+nharm])
        m = hat_hat[np.ix_(jj, jj)]
        dphase[j] = np.sqrt(np.dot(np.dot(v, m), v))

    out_dict['amplitude'] = amp
    out_dict['amplitude_error'] = damp
    out_dict['rel_phase'] = rel_phase
    out_dict['rel_phase_error'] = dphase
    out_dict['time0'] = -phase[0] / (2 * np.pi * freq)

    ncp = norm.cumprod()
    out_dict['trend_coef'] = coef / ncp
    out_dict['y_offset'] = out_dict['trend_coef'][0] - cn0

    prob = stats.f.sf(0.5 * (ntime - 1. - detrend_order) * (1. -out_dict['chi2'] / out_dict['chi0']), 2, ntime - 1 - detrend_order)
    out_dict['signif'] = lprob2sigma(np.log(prob))

    return out_dict


def get_lomb_frequency(lomb_model, i):
    """Get the ith frequency from a fitted Lomb-Scargle model."""
    return lomb_model['freq_fits'][i-1]['freq']


def get_lomb_amplitude(lomb_model, i, j):
    """
    Get the amplitude of the jth harmonic of the ith frequency from a fitted
    Lomb-Scargle model.
    """
    return lomb_model['freq_fits'][i-1]['amplitude'][j-1]


def get_lomb_rel_phase(lomb_model, i, j):
    """
    Get the relative phase of the jth harmonic of the ith frequency from a
    fitted Lomb-Scargle model.
    """
    return lomb_model['freq_fits'][i-1]['rel_phase'][j-1]


def get_lomb_amplitude_ratio(lomb_model, i):
    """
    Get the ratio of the amplitudes of the first harmonic for the ith and first
    frequencies from a fitted Lomb-Scargle model.
    """
    return (lomb_model['freq_fits'][i-1]['amplitude'][0] /
            lomb_model['freq_fits'][0]['amplitude'][0])


def get_lomb_frequency_ratio(lomb_model, i):
    """
    Get the ratio of the ith and first frequencies from a fitted Lomb-Scargle
    model.
    """
    return (lomb_model['freq_fits'][i-1]['freq'] /
            lomb_model['freq_fits'][0]['freq'])


def get_lomb_signif_ratio(lomb_model, i):
    """
    Get the ratio of the significances (in sigmas) of the ith and first
    frequencies from a fitted Lomb-Scargle model.
    """
    return (lomb_model['freq_fits'][i-1]['signif'] /
            lomb_model['freq_fits'][0]['signif'])


def get_lomb_lambda(lomb_model):
    """Get the regularization parameter of a fitted Lomb-Scargle model."""
    return lomb_model['freq_fits'][0]['lambda']


def get_lomb_signif(lomb_model):
    """
    Get the significance (in sigmas) of the first frequency from a fitted
    Lomb-Scargle model.
    """
    return lomb_model['freq_fits'][0]['signif']


def get_lomb_varrat(lomb_model):
    """
    Get the fraction of the variance explained by the first frequency of a
    fitted Lomb-Scargle model.
    """
    return lomb_model['varrat']


def get_lomb_trend(lomb_model):
    """Get the linear trend of a fitted Lomb-Scargle model."""
    return lomb_model['trend']


def get_lomb_y_offset(lomb_model):
    """Get the y-intercept of a fitted Lomb-Scargle model."""
    return lomb_model['freq_fits'][0]['y_offset']



from . import common_functions as cf


def scatter_res_raw(t, m, e, lomb_model):
    """ From arXiv 1101_2406v1 Dubath 20110112 paper.

    Scatter: res/raw
    Median absolute deviation (MAD) of the residuals (obtained by subtracting
    model values from the raw light curve) divided by the MAD of the raw
    light-curve values around the median.
    """
    lomb_resid = lomb_model['freq_fits'][-1]['resid']
    return (cf.median_absolute_deviation(lomb_resid) /
            cf.median_absolute_deviation(m))



import numpy as np
import gatspy


def lomb_scargle_fast_period(t, m, e):
    """Fits a simple sinuosidal model

        y(t) = A sin(2*pi*w*t + phi) + c

    and returns the estimated period 1/w. Much faster than fitting the
    full multi-frequency model used by `features.lomb_scargle`.
    """
    dt = t.max() - t.min()
    opt_args = {'period_range': (2 * dt / len(t), dt), 'quiet': True}
    model = gatspy.periodic.LombScargleFast(fit_period=True,
                                            optimizer_kwds=opt_args,
                                            silence_warnings=True)
    model.fit(t, m, e)
    return model.best_period



import itertools
import numpy as np
import numpy.testing as npt
from cesium.features import cadence_features as cf
from cesium.features.tests.util import irregular_random


def test_delta_t_hist():
    """Test histogram of all time lags."""
    times, values, errors = irregular_random(500)
    delta_ts = [pair[1] - pair[0] for pair in itertools.combinations(times, 2)]
    nbins = 50
    bins = np.linspace(0, max(times) - min(times), nbins + 1)
    npt.assert_allclose(cf.delta_t_hist(times, nbins),
                        np.histogram(delta_ts, bins=bins)[0], atol=2)


def test_normalize_hist():
    """Test normalization of histogram."""
    times, values, errors = irregular_random(500)
    delta_ts = [pair[1] - pair[0] for pair in itertools.combinations(times, 2)]
    nbins = 50
    bins = np.linspace(0, max(times) - min(times), nbins + 1)
    nhist = cf.normalize_hist(cf.delta_t_hist(times, nbins),
                              max(times) - min(times))
    npt.assert_allclose(nhist, np.histogram(delta_ts, bins=bins,
                                            density=True)[0], atol=0.01)


def test_find_sorted_peaks():
    """Test peak-finding algorithm."""
    x = np.array([0, 5, 3, 1])  # Single peak
    npt.assert_allclose(cf.find_sorted_peaks(x), np.array([[1, 5]]))

    x = np.array([0, 5, 3, 6, 1])  # Multiple peaks
    npt.assert_allclose(cf.find_sorted_peaks(x), np.array([[3, 6], [1, 5]]))

    x = np.array([3, 1, 3])  # End-points can be peaks
    npt.assert_allclose(cf.find_sorted_peaks(x), np.array([[0, 3], [2, 3]]))

    x = np.array([0, 3, 3, 3, 0])  # In case of ties, peak is left-most point
    npt.assert_allclose(cf.find_sorted_peaks(x), np.array([[1, 3]]))

    # Tie is a peak only if greater than next value
    x = np.array([0, 3, 3, 5, 0])
    npt.assert_allclose(cf.find_sorted_peaks(x), np.array([[3, 5]]))


def test_peak_ratio():
    """ Test peak ratio method."""
    x = np.array([0, 5, 2, 3, 1])
    peaks1 = cf.find_sorted_peaks(x)
    npt.assert_almost_equal(cf.peak_ratio(peaks1, 0, 1), 5 / 3)
    assert cf.peak_ratio(peaks1, 1, 6) is np.nan
    peaks2 = []
    assert cf.peak_ratio(peaks2, 1, 2) is np.nan


def test_peak_bins():
    """ Test peak bins method """
    x = np.array([0, 5, 2, 3, 1])
    peaks1 = cf.find_sorted_peaks(x)
    npt.assert_almost_equal(cf.peak_bin(peaks1, 0), 1)
    npt.assert_almost_equal(cf.peak_bin(peaks1, 1), 3)
    result1 = cf.peak_bin(peaks1, 6)
    assert cf.peak_bin(peaks1, 6) is np.nan



from _lomb_scargle cimport lomb_scargle as _lomb_scargle

cimport numpy as cnp
import numpy as np

def lomb_scargle(int numt, int numf, int nharm, int detrend_order,
                 double[:] psd, double[:] cn, cnp.ndarray wth,
                 double[:] sinx, double[:] cosx, double[:] sinx_step,
                 double[:] cosx_step, double[:] sinx_back,
                 double[:] cosx_back, double[:] sinx_smallstep,
                 double[:] cosx_smallstep, double[:, :] hat_matr,
                 double[:, :] hat_hat, double[:, :] hat0,
                 double[:] soln, double chi0, double freq_zoom,
                 double psdmin, double tone_control,
                 cnp.ndarray[dtype=double, ndim=0] lambda0,
                 double[:] lambda0_range,
                 cnp.ndarray[dtype=double, ndim=0] Tr,
                 cnp.ndarray[dtype=cnp.int32_t, ndim=0] ifreq):

    assert wth.dtype == np.double

    _lomb_scargle(numt, numf, nharm, detrend_order, &psd[0], &cn[0],
                  <double*>(wth.data), &sinx[0], &cosx[0], &sinx_step[0],
                  &cosx_step[0], &sinx_back[0], &cosx_back[0],
                  &sinx_smallstep[0], &cosx_smallstep[0], &hat_matr[0, 0],
                  &hat_hat[0, 0], &hat0[0, 0],
                  &soln[0], chi0, freq_zoom, psdmin, tone_control,
                  <double*>(lambda0.data), &lambda0_range[0],
                  <double*>(Tr.data), <int*>(ifreq.data))



import numpy as np
from scipy import optimize


# TODO what is this exactly?
def periodic_model(lomb_model):
    """
    Compute features related to the extreme points of the fitted Lomb Scargle
    model.
    """
    out_dict = {}

    A = lomb_model['freq_fits'][0]['amplitude']
    ph = lomb_model['freq_fits'][0]['rel_phase']

    def model_f(t):
        return (A[0] * np.sin(2. * np.pi      * t + ph[0]) +
                A[1] * np.sin(2. * np.pi * 2. * t + ph[1]) +
                A[2] * np.sin(2. * np.pi * 3. * t + ph[2]) +
                A[3] * np.sin(2. * np.pi * 4. * t + ph[3]) +
                A[4] * np.sin(2. * np.pi * 5. * t + ph[4]) +
                A[5] * np.sin(2. * np.pi * 6. * t + ph[5]) +
                A[6] * np.sin(2. * np.pi * 7. * t + ph[6]) +
                A[7] * np.sin(2. * np.pi * 8. * t + ph[7]))

    def model_neg(t):
        return -1. * model_f(t)

    # Start finding 1st minima, at 5% of phase (fudge/magic number) > 0.018
    min_1_a = optimize.fmin(model_neg, 0.05, disp=False)[0]
    max_2_a = optimize.fmin(model_f, min_1_a + 0.01, disp=False)[0]
    min_3_a = optimize.fmin(model_neg, max_2_a + 0.01, disp=False)[0]
    max_4_a = optimize.fmin(model_f, min_3_a + 0.01, disp=False)[0]

# TODO !!! is this wrong? seems like it should be a minus
    out_dict['phi1_phi2'] = (min_3_a - max_2_a) / (max_4_a / min_3_a)
    out_dict['min_delta_mags'] = abs(model_f(min_1_a) - model_f(min_3_a))
    out_dict['max_delta_mags'] = abs(model_f(max_2_a) - model_f(max_4_a))
    return out_dict


def get_max_delta_mags(model):
    """Largest value minus second largest value of fitted Lomb Scargle model."""
    return model['max_delta_mags']


def get_min_delta_mags(model):
    """Second smallest value minus smallest value of fitted Lomb Scargle model."""
    return model['min_delta_mags']


def get_model_phi1_phi2(model):
    """
    Ratio of distances between the second minimum and first maximum, and the
    second minimum and second maximum, of the fitted Lomb-Scargle model.
    """
    return model['phi1_phi2']



from tabulate import tabulate
import re
from cesium.features.graphs import (feature_categories, dask_feature_graph,
                                    extra_feature_docs)


def feature_graph_to_rst_table(graph, category_name):
    """Convert feature graph to Sphinx-compatible ReST table."""
    header = [category_name, 'Description']
    table = []
    for feature_name in sorted(
            graph, key=lambda s: [int(t) if t.isdigit() else t
                                  for t in re.split('(\d+)', s)]):
        description = (extra_feature_docs[feature_name] if feature_name in
                       extra_feature_docs else
                       ' '.join([e.strip() for e in
                                 graph[feature_name][0].__doc__.split('\n\n')[0]
                                 .strip().split('\n')]))
        table.append([feature_name, description])

    return tabulate(table, headers=header, tablefmt='rst')


def write_feature_tables(fname):
    with open(fname, 'w') as f:
        f.write('==============================\n'
                'Cesium Features - By Category\n'
                '==============================\n\n')

        dfg = dask_feature_graph

        for category in feature_categories:
            graph = {feature: dfg[feature]
                         for feature in feature_categories[category]}

            f.write(feature_graph_to_rst_table(graph, category) + '\n\n')



import numpy as np
import scipy.stats as stats


__all__ = ['double_to_single_step', 'cad_prob', 'delta_t_hist',
           'normalize_hist', 'find_sorted_peaks', 'peak_ratio', 'peak_bin']


def double_to_single_step(cads):
    """Ratios (t[i+2] - t[i]) / (t[i+1] - t[i])."""
    cads = np.asarray(cads)
    return (cads[2:] + cads[:-2]) / (cads[1:-1] - cads[:-2])


def cad_prob(cads, time):
    """Given the observed distribution of time lags `cads`, compute the probability
    that the next observation occurs within `time` minutes of an arbitrary epoch.
    """
    return stats.percentileofscore(cads, float(time) / (24.0 * 60.0)) / 100.0


def delta_t_hist(t, nbins=50, conv_oversample=50):
    """Build histogram of all possible |t_i - t_j|'s.

    For efficiency, we construct the histogram via a convolution of the PDF
    rather than by actually computing all the differences. For better accuracy
    we use a factor `conv_oversample` more bins when performing the convolution
    and then aggregate the result to have `nbins` total values.
    """
    f, x = np.histogram(t, bins=conv_oversample * nbins)
    g = np.convolve(f, f[::-1])[len(f) - 1:]  # Discard negative domain
    g[0] -= len(t)  # First bin is double-counted because of i=j terms
    hist = g.reshape((-1, conv_oversample)).sum(axis=1)  # Combine bins
    return hist


def normalize_hist(hist, total_time):
    """Normalize histogram such that integral from t_min to t_max equals 1.
    cf. np.histogram(..., density=True).
    """
    return hist / (total_time * np.mean(hist))


def find_sorted_peaks(x):
    """Find peaks, i.e. local maxima, of an array. Interior points are peaks if
    they are greater than both their neighbors, and edge points are peaks if
    they are greater than their only neighbor. In the case of ties, we
    (arbitrarily) choose the first index in the sequence of equal values as the
    peak.
    Returns a list of tuples (i, x[i]) of peak indices i and values x[i],
    sorted in decreasing order by peak value.
    """
    peak_inds = []
    nbins = len(x)
    for i in range(nbins):
        if i == 0 or x[i] > x[i - 1]:  # Increasing from left
            if i == nbins - 1 or x[i] > x[i + 1]:  # Increasing from right
                peak_inds.append(i)
            elif x[i] == x[i + 1]:  # Tied; check the next non-equal value
                for j in range(i + 1, nbins):
                    if x[j] != x[i]:
                        if x[j] < x[i]:
                            peak_inds.append(i)
                        break
                if j == nbins - 1 and x[i] == x[j]:  # Reached the end
                    peak_inds.append(i)
    sorted_peak_inds = sorted(peak_inds, key=lambda i: x[i], reverse=True)
    return list(zip(sorted_peak_inds, x[sorted_peak_inds]))


def peak_ratio(peaks, i, j):
    """Compute the ratio of the values of the ith and jth largest peaks. Peaks is a list
    of tuples (i, x[i]) of peak indices i and values x[i], sorted in decreasing order
    by peak value."""
    if len(peaks) > i and len(peaks) > j:
        return peaks[i][1] / peaks[j][1]
    else:
        return np.nan


def peak_bin(peaks, i):
    """Return the (bin) index of the ith largest peak. Peaks is a list of tuples (i, x[i])
    of peak indices i and values x[i], sorted in decreasing order by peak value."""
    if len(peaks) > i:
        return peaks[i][0]
    else:
        return np.nan



import numpy as np
from . import lomb_scargle as ls
from . import common_functions as cf


# TODO is this worth it since it doubles running time?
def period_folding(x, y, dy, lomb_model, sys_err=0.05):
    """
    This section is used to calculate Dubath (10. Percentile90:2P/P),
    which requires regenerating a model using 2P where P is the original found period

    NOTE: this essentially runs everything a second time, so makes feature
    generation take roughly twice as long.
    """
    out_dict = {}
    model_vals = np.zeros(len(y))
    freq_2p = lomb_model['freq_fits'][0]['freq'] * 0.5
    ytest_2p = y.copy()
    dy0 = np.sqrt(dy**2 + sys_err**2)

    # Here we force the freq to just 2*freq1_Period; we also do not use linear
    # detrending since we are not searching for freqs, and we want the
    # resulting model to be smooth when in phase-space. Detrending would result
    # in non-smooth model when period folded
    lambda0_range = [-np.log10(len(x)), 8.]
    fit = ls.fit_lomb_scargle(x, ytest_2p, dy0, freq_2p, lomb_model['df'], 1,
            lambda0_range=lambda0_range, nharm=lomb_model['nharm'], detrend_order=0)
    model_vals += fit['model']

    ytest_2p -= fit['model']
    for i in range(1, lomb_model['nfreq']):
        fit = ls.fit_lomb_scargle(x, ytest_2p, dy0, lomb_model['f0'],
                lomb_model['df'], lomb_model['numf'],
                lambda0_range=lambda0_range, nharm=lomb_model['nharm'],
                detrend_order=0)
        ytest_2p -= fit['model']

    out_dict['1p_resid'] = lomb_model['freq_fits'][-1]['resid']
    out_dict['2p_resid'] = ytest_2p

    # So the following uses the 2*Period model, and gets a time-sorted, folded t and m:
    # NOTE: if this is succesful, I think a lot of other features could characterize the
    # shapes of the 2P folded data (not P or 2P dependent).
    # the reason we choose 2P is that occasionally for eclipsing
    # sources the LS code chooses 0.5 of true period (but never 2x
    # the true period).  slopes are not dependent upon the actual
    # period so 2P is fine if it gives a high chance of correct fitting.
    # NOTE: we only use the model from freq1 because this with its harmonics seems to
    # adequately model shapes such as RRLyr skewed sawtooth, multi minima of rvtau
    # without getting the scatter from using additional LS found frequencies.
    t_2per_fold = np.array(x % (1. / freq_2p))
    t_2per_sort_inds = np.argsort(t_2per_fold)
    t_2per_fold = t_2per_fold[t_2per_sort_inds]
    y_2per_fold = np.array(model_vals)[t_2per_sort_inds]
    out_dict['folded_slopes'] = slopes = np.diff(y_2per_fold) / np.diff(t_2per_fold)

    return out_dict


def p2p_model(x, y, frequency):
    """
    Compute features that compare the residuals of data folded by estimated
    period from Lomb-Scargle model with residuals folded by twice the estimated
    period.
    """

    sumsqr_diff_unfold = np.sum((np.diff(y)**2))
    median_diff = np.median(np.abs(np.diff(y)))
    mad = cf.median_absolute_deviation(y)
    x = x.copy()
    x = x - min(x)

    t_2per_fold = np.array(x % (2. / frequency))
    t_2per_sort_inds = np.argsort(t_2per_fold)
    t_2per_fold = t_2per_fold[t_2per_sort_inds]
    y_2per_fold = np.array(y)[t_2per_sort_inds]
    sumsqr_diff_2per_fold = np.sum(np.diff(y_2per_fold)**2)

    ### eta feature from arXiv 1101.3316 Kim QSO paper:
    t_1per_fold = np.array(x % (1. / frequency))
    t_1per_sort_inds = np.argsort(t_1per_fold)
    t_1per_fold = t_1per_fold[t_1per_sort_inds]
    y_1per_fold = np.array(y)[t_1per_sort_inds]
    median_1per_fold_diff = np.median(np.abs(np.diff(y_1per_fold)))

    out_dict = {}
    out_dict['scatter_2praw'] = sumsqr_diff_2per_fold / sumsqr_diff_unfold
    out_dict['scatter_over_mad'] = median_diff / mad
    out_dict['ssqr_diff_over_var'] = sumsqr_diff_unfold / ((len(y) - 1)
            * np.var(y))
    out_dict['scatter_pfold_over_mad'] = median_1per_fold_diff / mad
    return out_dict


# TODO why not just get the (almost) steepest positive/negative slopes directly?
# TODO wrong for strictly increasing/decreasing
def get_fold2P_slope_percentile(model, alpha):
    """Get alphath percentile of slopes of period-folded model."""
    return np.percentile(model['folded_slopes'], alpha)


def get_medperc90_2p_p(model):
    """
    Get ratio of 90th percentiles of residuals for data folded by twice the
    estimated period and the estimated period, respectively.
    """
    return (np.percentile(np.abs(model['2p_resid']), 90) /
            np.percentile(np.abs(model['1p_resid']), 90))


def get_p2p_scatter_2praw(model):
    """
    Get ratio of variability (sum of squared differences of consecutive
    values) of folded and unfolded models.
    """
    return model['scatter_2praw']


def get_p2p_scatter_over_mad(model):
    """Get ratio of variability of folded and unfolded models."""
    return model['scatter_over_mad']


def get_p2p_scatter_pfold_over_mad(model):
    """
    Get ratio of median of period-folded data over median absolute
    deviation of observed values.
    """
    return model['scatter_pfold_over_mad']


def get_p2p_ssqr_diff_over_var(model):
    """
    Get sum of squared differences of consecutive values as a fraction of the
    variance of the data.
    """
    return model['ssqr_diff_over_var']



"""
Attempt to generate templates for module reference with Sphinx

Notes
-----
This parsing is based on import and introspection of modules.

This is a modified version of a script shipped with scikit-image, which was
itself based on similar scripts from PyMVPA and NiPy.
"""

# Stdlib imports
import os
import re

from types import BuiltinFunctionType, FunctionType

# suppress print statements (warnings for empty files)
DEBUG = True


class ApiDocWriter(object):
    ''' Class for automatic detection and parsing of API docs
    to Sphinx-parsable reST format'''

    # only separating first two levels
    rst_section_levels = ['*', '=', '-', '~', '^']

    def __init__(self,
                 package_name,
                 rst_extension='.txt',
                 skip_patterns=['.*__', '.*tests$'],
                 ):
        ''' Initialize package for parsing

        Parameters
        ----------
        package_name : string
            Name of the top-level package.  *package_name* must be the
            name of an importable package
        rst_extension : string, optional
            Extension for reST files, default '.rst'
        skip_patterns : None or sequence
            Sequence of strings giving URIs of modules to be excluded
            Operates on the module name including preceding URI path,
            back to the first dot after *package_name*.  For example
            ``sphinx.util.console`` results in the string to search of
            ``.util.console``
            If is None, gives default. Default is:
            ['\.setup$', '\._']
        '''
        self.package_name = package_name
        self.skip_patterns = skip_patterns
        self.rst_extension = rst_extension
        root_module = self._import(package_name)
        self.root_path = root_module.__path__[-1]

    def _import(self, name):
        ''' Import namespace package '''
        mod = __import__(name)
        components = name.split('.')
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod

    def _get_object_name(self, line):
        ''' Get second token in line
        >>> docwriter = ApiDocWriter('sphinx')
        >>> docwriter._get_object_name("  def func():  ")
        'func'
        >>> docwriter._get_object_name("  class Klass(object):  ")
        'Klass'
        >>> docwriter._get_object_name("  class Klass:  ")
        'Klass'
        '''
        name = line.split()[1].split('(')[0].strip()
        # in case we have classes which are not derived from object
        # ie. old style classes
        return name.rstrip(':')

    def _parse_module_with_import(self, uri):
        """Look for functions and classes in an importable module.

        Parameters
        ----------
        uri : str
            The name of the module to be parsed. This module needs to be
            importable.

        Returns
        -------
        functions : list of str
            A list of (public) function names in the module.
        classes : list of str
            A list of (public) class names in the module.
        """
        mod = __import__(uri, fromlist=[uri.split('.')[-1]])
        # find all public objects in the module.
        obj_strs = [obj for obj in dir(mod) if not obj.startswith('_')]
        functions = []
        classes = []
        for obj_str in obj_strs:
            # find the actual object from its string representation
            if obj_str not in mod.__dict__:
                continue
            obj = mod.__dict__[obj_str]
            if (hasattr(obj, '__module__')
                and self.package_name not in obj.__module__):
                continue

            # figure out if obj is a function or class
            if isinstance(obj, (FunctionType, BuiltinFunctionType)):
                functions.append(obj_str)
            else:
                try:
                    issubclass(obj, object)
                    classes.append(obj_str)
                except TypeError:
                    # not a function or class
                    pass
        return functions, classes

    def generate_api_doc(self, uri):
        '''Make autodoc documentation template string for a module

        Parameters
        ----------
        uri : string
            python location of module - e.g 'sphinx.builder'

        Returns
        -------
        S : string
            Contents of API doc
        '''
        # get the names of all classes and functions
        functions, classes = self._parse_module_with_import(uri)
        if not len(functions) and not len(classes) and DEBUG:
            print('WARNING: Empty -', uri)  # dbg
            return ''

        # Make a shorter version of the uri that omits the package name for
        # titles
        uri_short = re.sub(r'^%s\.' % self.package_name,'',uri)

        ad = '.. AUTO-GENERATED FILE -- DO NOT EDIT!\n\n'

        # Set the chapter title to read 'module' for all modules except for the
        # main packages
        if '.' in uri:
            title = 'Module: :mod:`' + uri_short + '`'
        else:
            title = ':mod:`' + uri_short + '`'
        ad += title + '\n' + self.rst_section_levels[1] * len(title)

        ad += '\n.. automodule:: ' + uri + '\n'
        ad += '\n.. currentmodule:: ' + uri + '\n'
        ad += '.. autosummary::\n\n'
        for f in functions:
            ad += '   ' + uri + '.' + f + '\n'
        ad += '\n'
        for c in classes:
            ad += '   ' + uri + '.' + c + '\n'
        ad += '\n'

        for f in functions:
            # must NOT exclude from index to keep cross-refs working
            full_f = uri + '.' + f
            ad += f + '\n'
            ad += self.rst_section_levels[2] * len(f) + '\n'
            ad += '\n.. autofunction:: ' + full_f + '\n\n'
        for c in classes:
            ad += '\n:class:`' + c + '`\n' \
                  + self.rst_section_levels[2] * \
                  (len(c)+9) + '\n\n'
            ad += '\n.. autoclass:: ' + c + '\n'
            # must NOT exclude from index to keep cross-refs working
            ad += '  :members:\n' \
                  '  :undoc-members:\n' \
                  '  :show-inheritance:\n' \
                  '\n' \
                  '  .. automethod:: __init__\n'
        return ad

    def _survives_exclude(self, matchstr):
        ''' Returns True if *matchstr* does not match patterns

        ``self.package_name`` removed from front of string if present
        '''
        for pat in self.skip_patterns:
            if re.compile(pat).search(matchstr):
                return False
        return True

    def _uri2path(self, uri):
        ''' Convert uri to absolute filepath

        Parameters
        ----------
        uri : string
            URI of python module to return path for

        Returns
        -------
        path : None or string
            Returns None if there is no valid path for this URI
            Otherwise returns absolute file system path for URI

        Examples
        --------
        >>> docwriter = ApiDocWriter('sphinx')
        >>> import sphinx
        >>> modpath = sphinx.__path__[0]
        >>> res = docwriter._uri2path('sphinx.builder')
        >>> res == os.path.join(modpath, 'builder.py')
        True
        >>> res = docwriter._uri2path('sphinx')
        >>> res == os.path.join(modpath, '__init__.py')
        True
        >>> docwriter._uri2path('sphinx.does_not_exist')

        '''
        if uri == self.package_name:
            return os.path.join(self.root_path, '__init__.py')
        path = uri.replace(self.package_name + '.', '')
        path = path.replace('.', os.path.sep)
        path = os.path.join(self.root_path, path)
        # XXX maybe check for extensions as well?
        if os.path.exists(path + '.py'): # file
            path += '.py'
        elif os.path.exists(os.path.join(path, '__init__.py')):
            path = os.path.join(path, '__init__.py')
        else:
            return None
        return path

    def _path2uri(self, dirpath):
        ''' Convert directory path to uri '''
        package_dir = self.package_name.replace('.', os.path.sep)
        relpath = dirpath.replace(self.root_path, package_dir)
        if relpath.startswith(os.path.sep):
            relpath = relpath[1:]
        return relpath.replace(os.path.sep, '.')

    def discover_modules(self):
        ''' Return module sequence discovered from ``self.package_name``


        Parameters
        ----------
        None

        Returns
        -------
        mods : sequence
            Sequence of module names within ``self.package_name``

        Examples
        --------
        >>> dw = ApiDocWriter('sphinx')
        >>> mods = dw.discover_modules()
        >>> 'sphinx.util' in mods
        True
        >>> dw.skip_patterns.append('\.util$')
        >>> 'sphinx.util' in dw.discover_modules()
        False
        >>>
        '''
        modules = [self.package_name]
        # raw directory parsing
        for dirpath, dirnames, filenames in os.walk(self.root_path,
                                                    topdown=False):
            # Check directory names for packages
            root_uri = self._path2uri(os.path.join(self.root_path,
                                                   dirpath))
            filenames = [f for f in filenames if f.endswith('.py')]
            uris = [u.strip('.py') for u in dirnames + filenames]
            for uri in uris:
                package_uri = '.'.join((root_uri, uri))
                if (self._uri2path(package_uri) and
                    self._survives_exclude(package_uri)):
                    try:
                        mod = __import__(package_uri, fromlist=['cesium'])
                        mod.__all__
                        modules.append(package_uri)
                    except (ImportError, AttributeError):
                        pass
        return sorted(modules)

    def write_modules_api(self, modules, outdir):
        # write the list
        written_modules = []
        for m in modules:
            api_str = self.generate_api_doc(m)
            if not api_str:
                continue
            # write out to file
            outfile = os.path.join(outdir,
                                   m + self.rst_extension)
            fileobj = open(outfile, 'wt')
            fileobj.write(api_str)
            fileobj.close()
            written_modules.append(m)
        self.written_modules = written_modules

    def write_api_docs(self, outdir):
        """Generate API reST files.

        Parameters
        ----------
        outdir : string
            Directory name in which to store files
            We create automatic filenames for each module

        Returns
        -------
        None

        Notes
        -----
        Sets self.written_modules to list of written modules
        """
        if not os.path.exists(outdir):
            os.mkdir(outdir)
        # compose list of modules
        modules = self.discover_modules()
        self.write_modules_api(modules, outdir)

    def write_index(self, outdir, froot='gen', relative_to=None):
        """Make a reST API index file from written files

        Parameters
        ----------
        path : string
            Filename to write index to
        outdir : string
            Directory to which to write generated index file
        froot : string, optional
            root (filename without extension) of filename to write to
            Defaults to 'gen'.  We add ``self.rst_extension``.
        relative_to : string
            path to which written filenames are relative.  This
            component of the written file path will be removed from
            outdir, in the generated index.  Default is None, meaning,
            leave path as it is.
        """
        if self.written_modules is None:
            raise ValueError('No modules written')
        # Get full filename path
        path = os.path.join(outdir, froot+self.rst_extension)
        # Path written into index is relative to rootpath
        if relative_to is not None:
            relpath = (outdir + os.path.sep).replace(relative_to + os.path.sep, '')
        else:
            relpath = outdir
        print("outdir: ", relpath)
        idx = open(path,'wt')
        w = idx.write
        w('.. AUTO-GENERATED FILE -- DO NOT EDIT!\n\n')

        title = "API Reference"
        w(title + "\n")
        w("=" * len(title) + "\n\n")
        w('.. toctree::\n\n')
        for f in self.written_modules:
            w('   %s\n' % os.path.join(relpath,f))
        idx.close()



import os
from os.path import join as pjoin
import numpy as np
import pandas as pd
import scipy.stats
import dask

from cesium import featurize
from cesium.tests.fixtures import (sample_values, sample_ts_files,
                                   sample_featureset)

import numpy.testing as npt
import pytest


DATA_PATH = pjoin(os.path.dirname(__file__), "data")
FEATURES_CSV_PATH = pjoin(DATA_PATH, "test_features_with_targets.csv")


def test_featurize_files_function(tmpdir):
    """Test featurize function for on-disk time series"""
    with sample_ts_files(size=4, labels=['A', 'B']) as ts_paths:
        fset, labels = featurize.featurize_ts_files(ts_paths,
                                                    features_to_use=["std_err"],
                                                    scheduler=dask.get)
    assert "std_err" in fset
    assert fset.shape == (4, 1)
    npt.assert_array_equal(labels, ['A', 'B', 'A', 'B'])


def test_featurize_time_series_single():
    """Test featurize wrapper function for single time series"""
    t, m, e = sample_values()
    features_to_use = ['amplitude', 'std_err']
    meta_features = {'meta1': 0.5}
    fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                           meta_features, scheduler=dask.get)
    assert fset['amplitude'].values.dtype == np.float64


def test_featurize_time_series_single_multichannel():
    """Test featurize wrapper function for single multichannel time series"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    features_to_use = ['amplitude', 'std_err']
    meta_features = {'meta1': 0.5}
    fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                           meta_features, scheduler=dask.get)
    assert ('amplitude', 0) in fset.columns
    assert 'meta1' in fset.columns


def test_featurize_time_series_multiple():
    """Test featurize wrapper function for multiple time series"""
    n_series = 5
    list_of_series = [sample_values() for i in range(n_series)]
    times, values, errors = [list(x) for x in zip(*list_of_series)]
    features_to_use = ['amplitude', 'std_err']
    meta_features = [{'meta1': 0.5}] * n_series
    fset = featurize.featurize_time_series(times, values, errors,
                                           features_to_use,
                                           meta_features, scheduler=dask.get)
    npt.assert_array_equal(sorted(fset.columns.get_level_values('feature')),
                           ['amplitude', 'meta1', 'std_err'])


def test_featurize_time_series_multiple_multichannel():
    """Test featurize wrapper function for multiple multichannel time series"""
    n_series = 5
    n_channels = 3
    list_of_series = [sample_values(channels=n_channels)
                      for i in range(n_series)]
    times, values, errors = [list(x) for x in zip(*list_of_series)]
    features_to_use = ['amplitude', 'std_err']
    meta_features = {'meta1': 0.5}
    fset = featurize.featurize_time_series(times, values, errors,
                                           features_to_use,
                                           meta_features, scheduler=dask.get)
    assert ('amplitude', 0) in fset.columns
    assert 'meta1' in fset.columns


def test_featurize_time_series_uneven_multichannel():
    """Test featurize wrapper function for uneven-length multichannel data"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    t = [[t, t[0:-5], t[0:-10]]]
    m = [[m[0], m[1][0:-5], m[2][0:-10]]]
    e = [[e[0], e[1][0:-5], e[2][0:-10]]]
    features_to_use = ['amplitude', 'std_err']
    meta_features = {'meta1': 0.5}
    fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                           meta_features, scheduler=dask.get)
    assert ('amplitude', 0) in fset.columns
    assert 'meta1' in fset.columns


def test_featurize_time_series_custom_functions():
    """Test featurize wrapper function for time series w/ custom functions"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    features_to_use = ['amplitude', 'std_err', 'test_f']
    meta_features = {'meta1': 0.5}
    custom_functions = {'test_f': lambda t, m, e: np.pi}
    fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                           meta_features,
                                           custom_functions=custom_functions,
                                           scheduler=dask.get)
    npt.assert_array_equal(fset['test_f', 0], np.pi)
    assert ('amplitude', 0) in fset.columns
    assert 'meta1' in fset.columns


def test_featurize_time_series_custom_dask_graph():
    """Test featurize wrapper function for time series w/ custom dask graph"""
    n_channels = 3
    t, m, e = sample_values(channels=n_channels)
    features_to_use = ['amplitude', 'std_err', 'test_f', 'test_meta']
    meta_features = {'meta1': 0.5}
    custom_functions = {'test_f': (lambda x: x.min() - x.max(), 'amplitude'),
                        'test_meta': (lambda x: 2. * x, 'meta1')}
    fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                           meta_features,
                                           custom_functions=custom_functions,
                                           scheduler=dask.get)
    assert ('amplitude', 0) in fset.columns
    assert ('test_f', 0) in fset.columns
    assert ('test_meta', 0) in fset.columns


def test_featurize_time_series_default_times():
    """Test featurize wrapper function for time series w/ missing times"""
    n_channels = 3
    _, m, e = sample_values(channels=n_channels)
    features_to_use = ['amplitude', 'std_err']
    meta_features = {}
    fset = featurize.featurize_time_series(None, m, e, features_to_use,
                                           meta_features, scheduler=dask.get)

    m = [[m[0], m[1][0:-5], m[2][0:-10]]]
    e = [[e[0], e[1][0:-5], e[2][0:-10]]]
    fset = featurize.featurize_time_series(None, m, e, features_to_use,
                                           meta_features, scheduler=dask.get)

    m = m[0][0]
    e = e[0][0]
    fset = featurize.featurize_time_series(None, m, e, features_to_use,
                                           meta_features, scheduler=dask.get)
    assert ('amplitude', 0) in fset.columns


def test_featurize_time_series_default_errors():
    """Test featurize wrapper function for time series w/ missing errors"""
    n_channels = 3
    t, m, _ = sample_values(channels=n_channels)
    features_to_use = ['amplitude', 'std_err']
    meta_features = {}
    fset = featurize.featurize_time_series(t, m, None, features_to_use,
                                           meta_features, scheduler=dask.get)

    t = [[t, t[0:-5], t[0:-10]]]
    m = [[m[0], m[1][0:-5], m[2][0:-10]]]
    fset = featurize.featurize_time_series(t, m, None, features_to_use,
                                           meta_features, scheduler=dask.get)

    t = t[0][0]
    m = m[0][0]
    fset = featurize.featurize_time_series(t, m, None, features_to_use,
                                           meta_features, scheduler=dask.get)
    assert ('amplitude', 0) in fset.columns


def test_featurize_time_series_pandas_metafeatures():
    """Test featurize function for metafeatures passed as Series/DataFrames."""
    t, m, e = sample_values()
    features_to_use = ['amplitude', 'std_err']
    meta_features = pd.Series({'meta1': 0.5})
    fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                           meta_features, scheduler=dask.get)
    npt.assert_allclose(fset['meta1'], 0.5)

    n_series = 5
    list_of_series = [sample_values() for i in range(n_series)]
    times, values, errors = [list(x) for x in zip(*list_of_series)]
    features_to_use = ['amplitude', 'std_err']
    meta_features = pd.DataFrame({'meta1': [0.5] * n_series,
                                  'meta2': [0.8] * n_series})
    fset = featurize.featurize_time_series(times, values, errors,
                                           features_to_use,
                                           meta_features, scheduler=dask.get)
    npt.assert_allclose(fset['meta1'], 0.5)
    npt.assert_allclose(fset['meta2'], 0.8)


def test_impute():
    """Test imputation of missing Featureset values."""
    fset, labels = sample_featureset(5, 1, ['amplitude'], ['class1', 'class2'],
                                     names=['a', 'b', 'c', 'd', 'e'],
                                     meta_features=['meta1'])

    imputed = featurize.impute_featureset(fset)
    npt.assert_allclose(fset.amplitude.values, imputed.amplitude.values)
    assert isinstance(imputed, pd.DataFrame)

    fset.amplitude.values[0] = np.inf
    fset.amplitude.values[1] = np.nan
    amp_values = fset.amplitude.values[2:]
    other_values = fset.values.T.ravel()[2:]

    imputed = featurize.impute_featureset(fset, strategy='constant',
                                          value=None)
    npt.assert_allclose(-2 * np.nanmax(np.abs(other_values)),
                        imputed.amplitude.values[0:2])

    imputed = featurize.impute_featureset(fset, strategy='constant',
                                          value=-1e4)
    npt.assert_allclose(-1e4, imputed.amplitude.values[0:2])

    imputed = featurize.impute_featureset(fset, strategy='mean')
    npt.assert_allclose(np.mean(amp_values), imputed.amplitude.values[0:2])
    npt.assert_allclose(amp_values, imputed.amplitude.values[2:])

    imputed = featurize.impute_featureset(fset, strategy='median')
    npt.assert_allclose(np.median(amp_values), imputed.amplitude.values[0:2])
    npt.assert_allclose(amp_values, imputed.amplitude.values[2:])

    imputed = featurize.impute_featureset(fset, strategy='most_frequent')
    npt.assert_allclose(scipy.stats.mode(amp_values).mode.item(),
                        imputed.amplitude.values[0:2])
    npt.assert_allclose(amp_values, imputed.amplitude.values[2:])

    featurize.impute_featureset(fset, strategy='constant', value=-1e4,
                                inplace=True)
    npt.assert_allclose(-1e4, fset.amplitude.values[0:2])

    with pytest.raises(NotImplementedError):
        featurize.impute_featureset(fset, strategy='blah')


def test_roundtrip_featureset(tmpdir):
    fset_path = os.path.join(str(tmpdir), 'test.npz')
    for n_channels in [1, 3]:
        for labels in [['class1', 'class2'], []]:
            fset, labels = sample_featureset(3, n_channels, ['amplitude'],
                                             labels, names=['a', 'b', 'c'],
                                             meta_features=['meta1'])

            pred_probs = pd.DataFrame(np.random.random((len(fset), 2)),
                                      index=fset.index.values,
                                      columns=['class1', 'class2'])

            featurize.save_featureset(fset, fset_path, labels=labels,
                                      pred_probs=pred_probs)
            fset_loaded, data_loaded = featurize.load_featureset(fset_path)
            npt.assert_allclose(fset.values, fset_loaded.values)
            npt.assert_array_equal(fset.index, fset_loaded.index)
            npt.assert_array_equal(fset.columns, fset_loaded.columns)
            assert isinstance(fset_loaded, pd.DataFrame)
            npt.assert_array_equal(labels, data_loaded['labels'])
            npt.assert_allclose(pred_probs, data_loaded['pred_probs'])
            npt.assert_array_equal(pred_probs.columns,
                                   data_loaded['pred_probs'].columns)


def test_ignore_exceptions():
    import cesium.features.graphs
    def raise_exc(x):
        raise ValueError()
    old_value = cesium.features.graphs.dask_feature_graph['mean']
    try:
        cesium.features.graphs.dask_feature_graph['mean'] = (raise_exc, 't')
        t, m, e = sample_values()
        features_to_use = ['mean']
        with pytest.raises(ValueError):
            fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                                   scheduler=dask.get,
                                                   raise_exceptions=True)
        fset = featurize.featurize_time_series(t, m, e, features_to_use,
                                               scheduler=dask.get,
                                               raise_exceptions=False)
        assert np.isnan(fset.values).all()
    finally:
        cesium.features.graphs.dask_feature_graph['mean'] = old_value



import contextlib
import errno
import os
import tarfile
import tempfile
import zipfile


__all__ = ['shorten_fname', 'remove_files', 'extract_time_series']


def shorten_fname(file_path):
    """Extract the name of a file (omitting directory names and extensions).

    Parameters
    ----------
    file_path : str
        Absolute or relative path to a file.

    Returns
    -------
    str
       The name of the file with directory names and extensions removed.

    """
    return os.path.splitext(os.path.basename(file_path))[0]


def remove_files(paths):
    """Remove specified file(s) from disk.

    Parameters
    ----------
    paths : str or list of str
        Path(s) to file(s) to be removed from disk.

    """
    if isinstance(paths, str):
        paths = [paths]

    for path in paths:
        try:
            os.remove(path)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
            else:
                pass


@contextlib.contextmanager
def extract_time_series(data_path, cleanup_archive=True, cleanup_files=False,
                        extract_dir=None):
    """Extract zip- or tarfile of time series file and return file paths.

    If the given file is not a tar- or zipfile then it is treated as a single
    time series filepath.

    Parameters
    ----------
    data_path : str
        Path to data archive or single data file.

    cleanup_archive : bool, optional
        Boolean specifying whether to delete the original archive (if
        applicable). Defaults to True.

    cleanup_files : bool, optional
        Boolean specifying whether to delete the extracted files when exiting
        the given context. Defaults to False.

    extract_dir : str, optional
        Directory into which files are to be extracted (if applicable). If
        None, a temporary directory is created.

    Yields
    ------
    list of str
        List of full paths to time series files.
    """
    if extract_dir is None:
        extract_dir = tempfile.mkdtemp()

    if tarfile.is_tarfile(data_path):
        archive = tarfile.open(data_path)
        members_to_extract = [x for x in archive.getmembers() if not
                              x.name.startswith(('.', '/'))]
        extracted_names = [x.name for x in members_to_extract]
        archive.extractall(path=extract_dir, members=members_to_extract)
        all_paths = [os.path.join(extract_dir, f) for f in extracted_names]
    elif zipfile.is_zipfile(data_path):
        archive = zipfile.ZipFile(data_path)
        members_to_extract = [x for x in archive.namelist() if not
                              x.startswith(('.', '/'))]
        archive.extractall(path=extract_dir, members=members_to_extract)
        all_paths = [os.path.join(extract_dir, f) for f in members_to_extract]
    else:
        archive = None
        all_paths = [data_path]

    if archive:
        archive.close()
        if cleanup_archive:
            remove_files(data_path)

    file_paths = [f for f in all_paths if not os.path.isdir(f)]
    try:
        yield file_paths
    finally:
        if cleanup_files:
            remove_files(file_paths)



import os
from uuid import uuid4
import numpy.testing as npt
import numpy as np
from cesium import time_series
from cesium.time_series import TimeSeries


def sample_time_series(size=51, channels=1):
    times = np.array([np.sort(np.random.random(size))
                      for i in range(channels)]).squeeze()
    values = np.array([np.random.normal(size=size)
                       for i in range(channels)]).squeeze()
    errors = np.array([np.random.exponential(size=size)
                       for i in range(channels)]).squeeze()
    return times, values, errors


def test__compatible_shapes():
    compat = time_series._compatible_shapes

    assert compat(np.arange(5), np.arange(5))
    assert not compat(np.arange(5), np.arange(6))

    assert compat([np.arange(5)] * 5, [np.arange(5)] * 5)
    assert not compat([np.arange(5)] * 5, [np.arange(5)] * 6)
    assert not compat([np.arange(5)] * 5, [np.arange(6)] * 5)
    assert not compat(np.arange(5), [np.arange(6)] * 5)

    assert compat([[0, 1], [0, 1]], [[0, 1], [0, 1]])
    assert not compat([[0, 1], [0, 1]], [[0], [0, 1]])

    assert compat([0, 1], np.arange(2))


def assert_ts_equal(ts1, ts2):
    for x1, x2 in zip((ts1.time, ts1.measurement, ts1.error),
                      (ts2.time, ts2.measurement, ts2.error)):
        assert type(x1) == type(x2)
        if isinstance(x1, np.ndarray):
            assert np.array_equal(x1, x2)
        else:
            assert all(np.array_equal(x1_i, x2_i)
                       for x1_i, x2_i in zip(x1, x2))
    assert ts1.label == ts2.label
    assert ts1.meta_features == ts2.meta_features
    assert ts1.name == ts2.name


def test_time_series_init_1d():
    t, m, e = sample_time_series(channels=1)
    ts = TimeSeries(t, m, e)
    assert ts.time.shape == t.shape and np.allclose(ts.time, t)
    assert ts.measurement.shape == m.shape and np.allclose(ts.measurement, m)
    assert ts.error.shape == e.shape and np.allclose(ts.error, e)
    assert ts.n_channels == 1


def test_time_series_init_2d():
    n_channels = 3
    t, m, e = sample_time_series(channels=n_channels)
    ts = TimeSeries(t, m, e)
    assert ts.time.shape == t.shape and np.allclose(ts.time, t)
    assert ts.measurement.shape == m.shape and np.allclose(ts.measurement, m)
    assert ts.error.shape == e.shape and np.allclose(ts.error, e)
    assert ts.n_channels == n_channels

    ts = TimeSeries(t[0], m, e[0])
    assert ts.time.shape == m.shape and np.allclose(ts.time[0], t[0])
    assert ts.measurement.shape == m.shape and np.allclose(ts.measurement, m)
    assert ts.error.shape == m.shape and np.allclose(ts.error[0], e[0])
    assert ts.n_channels == n_channels


def test_time_series_init_ragged():
    n_channels = 3
    t, m, e = sample_time_series(channels=n_channels)
    t = [t[i][0:i+2] for i in range(len(t))]
    m = [m[i][0:i+2] for i in range(len(m))]
    e = [e[i][0:i+2] for i in range(len(e))]
    ts = TimeSeries(t, m, e)
    assert all(np.allclose(ts.time[i], t[i]) for i in range(len(t)))
    assert all(np.allclose(ts.measurement[i], m[i]) for i in range(len(t)))
    assert all(np.allclose(ts.error[i], e[i]) for i in range(len(t)))
    assert ts.n_channels == n_channels


def test_time_series_default_values():
    n_channels = 3
    t, m, e = sample_time_series(channels=n_channels)
    ts = TimeSeries(None, m[0], None)
    npt.assert_allclose(ts.time,
                        np.linspace(0., time_series.DEFAULT_MAX_TIME,
                                    m.shape[1]))
    npt.assert_allclose(ts.error,
                        np.repeat(time_series.DEFAULT_ERROR_VALUE,
                                  m.shape[1]))
    assert ts.n_channels == 1

    ts = TimeSeries(None, m, None)
    npt.assert_allclose(ts.time[0],
                        np.linspace(0., time_series.DEFAULT_MAX_TIME,
                                    m.shape[1]))
    npt.assert_allclose(ts.error[0],
                        np.repeat(time_series.DEFAULT_ERROR_VALUE,
                                  m.shape[1]))
    assert ts.n_channels == n_channels

    t = [t[i][0:i+2] for i in range(len(t))]
    m = [m[i][0:i+2] for i in range(len(m))]
    e = [e[i][0:i+2] for i in range(len(e))]
    ts = TimeSeries(None, m, None)
    for i in range(n_channels):
        npt.assert_allclose(ts.time[i],
                            np.linspace(0., time_series.DEFAULT_MAX_TIME,
                                        len(m[i])))
        npt.assert_allclose(ts.error[i],
                            np.repeat(time_series.DEFAULT_ERROR_VALUE,
                                      len(m[i])))
    assert ts.n_channels == n_channels


def test_channels_iterator():
    n_channels = 3
    t, m, e = sample_time_series(channels=n_channels)
    ts = TimeSeries(t[0], m[0], e[0])
    for t_i, m_i, e_i in ts.channels():
        npt.assert_allclose(t_i, t[0])
        npt.assert_allclose(m_i, m[0])
        npt.assert_allclose(e_i, e[0])

    ts = TimeSeries(t, m, e)
    for (t_i, m_i, e_i), i in zip(ts.channels(), range(n_channels)):
        npt.assert_allclose(t_i, t[i])
        npt.assert_allclose(m_i, m[i])
        npt.assert_allclose(e_i, e[i])

    t = [t[i][0:i+2] for i in range(len(t))]
    m = [m[i][0:i+2] for i in range(len(m))]
    e = [e[i][0:i+2] for i in range(len(e))]
    ts = TimeSeries(t, m, e)
    for (t_i, m_i, e_i), i in zip(ts.channels(), range(n_channels)):
        npt.assert_allclose(t_i, t[i])
        npt.assert_allclose(m_i, m[i])
        npt.assert_allclose(e_i, e[i])


def test_time_series_npz(tmpdir):
    n_channels = 3
    t, m, e = sample_time_series(channels=n_channels)

    ts = TimeSeries(t[0], m[0], e[0])
    ts_path = os.path.join(str(tmpdir), str(uuid4()) + '.npz')
    ts.save(ts_path)
    ts_loaded = time_series.load(ts_path)
    assert_ts_equal(ts, ts_loaded)

    ts = TimeSeries(t[0], m, e[0])
    ts_path = os.path.join(str(tmpdir), str(uuid4()) + '.npz')
    ts.save(ts_path)
    ts_loaded = time_series.load(ts_path)
    assert_ts_equal(ts, ts_loaded)

    t = [t[i][0:i+2] for i in range(len(t))]
    m = [m[i][0:i+2] for i in range(len(m))]
    e = [e[i][0:i+2] for i in range(len(e))]
    ts = TimeSeries(t, m, e)
    ts_path = os.path.join(str(tmpdir), str(uuid4()) + '.npz')
    ts.save(ts_path)
    ts_loaded = time_series.load(ts_path)
    assert_ts_equal(ts, ts_loaded)


def test_time_series_sort():
    t, m, e = sample_time_series(channels=1)
    t[:2] = t[1::-1]
    ts = TimeSeries(t, m, e)
    npt.assert_allclose(ts.time, np.sort(t))
    npt.assert_allclose(ts.measurement, m[np.argsort(t)])
    npt.assert_allclose(ts.error, e[np.argsort(t)])

    n_channels = 3
    t, m, e = sample_time_series(channels=n_channels)
    t[:, :2] = t[:, 1::-1]
    ts = TimeSeries(t, m, e)
    for i in range(len(m)):
        npt.assert_allclose(ts.time[i], np.sort(t[i]))
        npt.assert_allclose(ts.measurement[i], m[i][np.argsort(t[i])])
        npt.assert_allclose(ts.error[i], e[i][np.argsort(t[i])])

    ts = TimeSeries(t[0], m, e[0])
    for i in range(len(m)):
        npt.assert_allclose(ts.time[i], np.sort(t[0]))
        npt.assert_allclose(ts.measurement[i], m[i][np.argsort(t[0])])
        npt.assert_allclose(ts.error[i], e[0][np.argsort(t[0])])



import dask
import numpy as np

from cesium.features import generate_dask_graph


def generate_features(t, m, e, features_to_use):
    """Utility function that generates features from a dask DAG."""
    graph = generate_dask_graph(t, m, e)
    values = dask.get(graph, features_to_use)
    return dict(zip(features_to_use, values))


def irregular_random(seed=0, size=50):
    """Generate random test data at irregularly-sampled times."""
    state = np.random.RandomState(seed)
    times = np.sort(state.uniform(0, 10, size))
    values = state.normal(1, 1, size)
    errors = state.exponential(0.1, size)
    return times, values, errors


def regular_periodic(freqs, amplitudes, phase, size=501):
    """Generate periodic test data sampled at regular intervals: superposition
    of multiple sine waves, each with multiple harmonics.
    """
    times = np.linspace(0, 2, size)
    values = np.zeros(size)
    for (i,j), amplitude in np.ndenumerate(amplitudes):
        values += amplitude * np.sin(2*np.pi*times*freqs[i]*(j+1) + phase)
    errors = 1e-4*np.ones(size)
    return times, values, errors


def irregular_periodic(freqs, amplitudes, phase, seed=0, size=501):
    """Generate periodic test data sampled at randomly-spaced intervals:
    superposition of multiple sine waves, each with multiple harmonics.
    """
    state = np.random.RandomState(seed)
    times = np.sort(state.uniform(0, 2, size))
    values = np.zeros(size)
    for i in range(freqs.shape[0]):
        for j in range(amplitudes.shape[1]):
            values += amplitudes[i,j] * np.sin(2*np.pi*times*freqs[i]*(j+1)+phase)
    errors = state.exponential(1e-2, size)
    return times, values, errors





import logging
import os

import numpy as np
import pandas as pd
from sklearn.externals import joblib

from .. import util
from .. import data_management
from . import util as dsutil


# TODO replace with real host
BASE_URL = "https://github.com/cesium-ml/cesium-data/raw/master/asas_training/"
MD5SUMS = {"asas_training_set.tar.gz": "02c65e90d23999ec1c59ad56a78de477"}
ARCHIVE_NAME = "asas_training_set.tar.gz"
HEADER_FILE = "asas_training_set_classes_with_metadata.dat"
CACHE_NAME = "asas_training_set.pkl"

logger = logging.getLogger(__name__)


def download_asas_training(data_dir):
    """Download sample light curve data.

    Three files are created within `data_dir`:
        - asas_training.tar.gz (containing .dat files for each time series)
        - asas_training.csv (header file containing class labels)
        - asas_training.pkl (cached data for faster loading from disk)

     Meta-features `meta1`, `meta2`, and `meta3` are randomly generated for
     testing purposes only.

    Parameters
    ----------
    data_dir: str
        Path where downloaded data should be stored.

    Returns
    -------
    dict
        Dictionary with attributes:
            - times: list of arrays of time values
            - measurements: list of arrays of measurement values
            - errors: list of arrays of error values
            - classes: array of class labels for each time series
            - metadata: DataFrame of metafeature values indexed by file
            - archive: path to data archive
            - header: path to header file
    """
    logger.warning("Downloading data from {}".format(BASE_URL))

    header_path = dsutil.download_file(data_dir, BASE_URL, HEADER_FILE)
    ts_paths = dsutil.download_and_extract_archives(data_dir, BASE_URL,
                                                    [ARCHIVE_NAME], MD5SUMS,
                                                    remove_archive=False)
    archive_path = os.path.join(data_dir, ARCHIVE_NAME)

    times = []
    measurements = []
    errors = []
    for fname in ts_paths:
        t, m, e = data_management.parse_ts_data(fname)
        times.append(t)
        measurements.append(m)
        errors.append(e)
    header = pd.read_csv(header_path, dtype={'filename': str})
    extract_dir = os.path.join(data_dir, os.path.basename(ARCHIVE_NAME))
    util.remove_files(ts_paths)

    classes, metadata = data_management.parse_headerfile(header_path, ts_paths)

    cache_path = os.path.join(data_dir, CACHE_NAME)
    data = dict(times=times, measurements=measurements, errors=errors,
                classes=classes, metadata=metadata, archive=archive_path,
                header=header_path)
    joblib.dump(data, cache_path, compress=3)
    return data


def fetch_asas_training(data_dir=None):
    """Download (if not already downloaded) and load example light curve data.

    Parameters
    ----------
    data_dir: str, optional
        Path where downloaded data should be stored. Defaults to
        a subdirectory `datasets/asas_training` within `dsutil.DATA_PATH`.

    Returns
    -------
    dict
        Dictionary attributes:
            - times: list of arrays of time values
            - measurements: list of arrays of measurement values
            - errors: list of arrays of error values
            - classes: Series of classes for each time series indexed by file
            - metadata: DataFrame of metafeature values indexed by file
            - archive: path to data archive
            - header: path to header file

    References
    ----------
    Richards, J. W., Starr, D. L., Miller, A. a., Bloom, J. S., Butler, N. R.,
    Brink, H., & Crellin-Quick, A. (2012). Construction of a Calibrated
    Probabilistic Classification Catalog: Application to 50k Variable Sources
    in the All-Sky Automated Survey. The Astrophysical Journal Supplement
    Series, 203(2), 32. http://doi.org/10.1088/0067-0049/203/2/32
    """

    if data_dir is None:
        data_dir = os.path.join(dsutil.DATA_PATH, "datasets/asas_training")
    cache_path = os.path.join(data_dir, CACHE_NAME)

    try:
        data = joblib.load(cache_path)
        logger.warning("Loaded data from cached archive.")
    except (ValueError, IOError): #  missing or incompatible cache
        data = download_asas_training(data_dir)
    return data
